import os
import time
import logging
from collections import OrderedDict
from typing import Optional, Sequence

import numpy
from fabio.edfimage import EdfImage
from silx.io import dictdump

from .items import Hdf5ItemWriter
from ..io import nexus
from ..io import h5_external
from ..io import h5_config
from ..io.io_utils import mkdir
from ..utils.array_order import Order
from ..utils.array_size import dtype_nbytes
from ..utils.array_size import shape_to_size
from ..utils.array_size import format_bytes
from ..nexus.mapping import normalize_nexus_name

logger = logging.getLogger(__name__)


class FileSizeMonitor:
    def __init__(self, filename="", timeout=10):
        self.filename = filename
        self.timeout = timeout

    def is_growing(self):
        filesize = os.path.getsize(self.filename)
        if filesize == self.filesize:
            if (time.time() - self.t0) > self.timeout:
                return False
        else:
            self.reset(filesize)
        return True

    def reset(self, filesize=0):
        self.filesize = filesize
        self.t0 = time.time()

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value
        self.reset()


class Hdf5DatasetWriter(Hdf5ItemWriter):
    """Manages creation and growth of one HDF5 dataset."""

    def __init__(
        self,
        filename=None,
        parent=None,
        filecontext=None,
        device=None,
        scan_shape=None,
        scan_save_shape=None,
        detector_shape=None,
        external_images_per_file=None,
        external_uri_from_file=False,
        dtype=None,
        chunk_options=None,
        saveorder=None,
        publishorder=None,
        parent_logger=None,
    ):
        """
        :param str filename: HDF5 file name
        :param filecontext: HDF5 open context manager
        :param str parent: path in the HDF5 file (must already exist)
        :param dict device: defined in module `writer.devices`
        :param tuple scan_shape: -1 indicate variable length
        :param tuple scan_save_shape: -1 indicate variable length
        :param tuple detector_shape:
        :param int external_images_per_file: number of images per file for external datasets
        :param bool external_uri_from_file: get the URI's from file instead of trusting the provided URI's
        :param dtype dtype:
        :param str compression_scheme: used whenever data is large enough to be chunked+compressed
        :param Order saveorder: order in which the scan shape is filled
        :param Order publishorder: order in which the scan shape is published
        :param parent_logger:
        """
        if parent_logger is None:
            parent_logger = logger
        super().__init__(
            filename=filename,
            parent=parent,
            filecontext=filecontext,
            parent_logger=parent_logger,
        )

        # Expected data shape, dtype and order
        if sum(n == 0 for n in scan_shape) > 1:
            raise ValueError("Scan can have only one variable dimension")
        if scan_save_shape is None:
            scan_save_shape = scan_shape
        elif sum(n == 0 for n in scan_save_shape) > 1:
            raise ValueError("Scan can have only one variable dimension")
        self._scan_shape = scan_shape
        self._scan_save_shape = scan_save_shape
        self._detector_shape = detector_shape
        self._ndim = len(scan_save_shape) + len(detector_shape)
        # Assume all dimensions are variable because:
        #   - channels can for example produce 2*n+1 data points where n the scan points
        #   - detectors can change dimensions during a scan (number of dimensions stays the same)
        self._maxshape = (None,) * self._ndim
        self._dtype = dtype
        if not isinstance(saveorder, Order):
            saveorder = Order(saveorder)
        self._saveorder = saveorder
        if not isinstance(publishorder, Order):
            publishorder = Order(publishorder)
        self._publishorder = publishorder

        # Derived from expected data shape, dtype and order
        self._itemsize = dtype_nbytes(dtype)
        if chunk_options is None:
            chunk_options = dict()
        self._h5_dset_config = h5_config.guess_dataset_config(
            scan_shape=scan_save_shape,
            detector_shape=detector_shape,
            dtype=dtype,
            max_shape=self._maxshape,
            **chunk_options,
        )
        if self._h5_dset_config["chunks"]:
            self._npoints_h5chunk = shape_to_size(
                self._h5_dset_config["chunks"][: self.scan_ndim]
            )
        else:
            self._npoints_h5chunk = 1
        self._save_interal_time = None
        self._save_interal_dtmax = 3

        # Currently arrive data shape (but not necessarily saved already)
        self.current_scan_save_shape = scan_save_shape
        self.current_detector_shape = detector_shape

        # Device parameters
        self._device = device

        # Internal/external data buffers
        self._internal_buffer = []  # Buffer data saved as an HDF5 dataset
        self._external_raw = []  # URI's for supported external binary data
        self._external_raw_formats = ["edf"]
        self._external_names = []  # URI's for unsupported external binary data
        self._external_datasets = []  # URI's for virtual datasets

        # External data settings
        self._external_images_per_file = external_images_per_file
        self._external_uri_from_file = external_uri_from_file

    @property
    def scan_shape(self):
        """Expected scan shape (when run to completion). -1 indicates
        a variable dimension.
        """
        return self._scan_shape

    @property
    def scan_save_shape(self):
        """Expected scan shape (when run to completion) as saved. -1
        indicates a variable dimension.
        """
        return self._scan_save_shape

    @property
    def detector_shape(self):
        """Expected detector shape. Does not contain -1 to indicate
        variable dimensions.
        """
        return self._detector_shape

    @property
    def dtype(self):
        """Data dtype"""
        return self._dtype

    @property
    def saveorder(self):
        """Order to fill `scan_save_shape` with data"""
        return self._saveorder

    @property
    def publishorder(self):
        """Order in which data from `scan_shape` arrives"""
        return self._publishorder

    @property
    def device(self):
        """Device parameters"""
        return self._device

    @property
    def external_images_per_file(self):
        """Number of images per file for external datasets (VDS)"""
        return self._external_images_per_file

    @property
    def external_uri_from_file(self):
        """Get the URI's from file instead of trusting the provided URI's"""
        return self._external_uri_from_file

    def __repr__(self):
        return "{}: shape = {}, dtype={}".format(
            repr(self.path), self.shape, self.dtype_name
        )

    @property
    def dtype_name(self):
        try:
            return self.dtype.__name__
        except AttributeError:
            return str(self.dtype)

    @property
    def name(self) -> str:
        return normalize_nexus_name(self.device["data_name"])

    @property
    def alias(self) -> Optional[str]:
        alias = self.device["data_alias"]
        if alias:
            alias = normalize_nexus_name(alias)
            if alias != self.name:
                return alias

    @property
    def linkname(self) -> Optional[str]:
        if self._external_names:
            # This dataset is a list of URI's saved as an array of strings
            return None
        return normalize_nexus_name(self.device["unique_name"])

    @property
    def device_type(self):
        device_type = self.device["device_type"]
        if not device_type:
            return f"unknown{self.detector_ndim}D"
        return device_type

    @property
    def data_type(self):
        return self.device["data_type"]

    @property
    def master_index(self):
        return self.device["master_index"]

    @property
    def scan_ndim(self):
        """Expected scan dimensions"""
        return len(self.scan_shape)

    @property
    def detector_ndim(self):
        """Expected detector dimensions"""
        return len(self.detector_shape)

    @property
    def scan_save_ndim(self):
        return len(self.scan_save_shape)

    @property
    def csaveorder(self):
        return self.saveorder.corder

    @property
    def cpublishorder(self):
        return self.publishorder.corder

    @property
    def shape(self):
        """Expected data shape. -1 indicates a variable length."""
        if self.csaveorder:
            return self.scan_save_shape + self.detector_shape
        else:
            return self.detector_shape + self.scan_save_shape

    @property
    def ndim(self):
        return self._ndim

    @property
    def current_shape(self):
        """Current data shape. -1 indicates a variable length."""
        if self.csaveorder:
            return self.current_scan_save_shape + self.current_detector_shape
        else:
            return self.current_detector_shape + self.current_scan_save_shape

    @property
    def current_h5shape(self):
        return tuple(max(n, 0) for n in self.current_shape)

    @property
    def current_scan_save_h5shape(self):
        return tuple(max(n, 0) for n in self.current_scan_save_shape)

    @property
    def grid_shape(self):
        """Like `current_shape` but with the original scan shape"""
        if self.variable_scan_shape and self.scan_ndim == 1:
            scan_shape = self.current_scan_save_shape
        else:
            scan_shape = self.scan_shape
        if self.csaveorder:
            return scan_shape + self.current_detector_shape
        else:
            return self.current_detector_shape + scan_shape

    @property
    def flat_shape(self):
        """Like `current_shape` but flatten the scan dimensions"""
        if self.scan_ndim:
            size = shape_to_size(self.current_scan_save_shape)
            if self.csaveorder:
                return (size,) + self.current_detector_shape
            else:
                return self.current_detector_shape + (size,)
        else:
            return self.current_detector_shape

    @property
    def reshaped(self):
        """
        Shape of scan equal to saved shape of scan?
        """
        return self.scan_shape != self.scan_save_shape

    @property
    def npoints_expected(self):
        # TODO: A fixed length scan can have channels with
        #       more than the expected points.
        return 0
        # return shape_to_size(self.scan_shape)

    @property
    def variable_shape(self):
        return any(n <= 0 for n in self.shape)

    @property
    def variable_scan_shape(self):
        return any(n <= 0 for n in self.scan_shape)

    @property
    def variable_detector_shape(self):
        # Variable detector shapes are not set to -1 in Redis
        # but to 1 so we will have to assume variable
        return True

    @property
    def current_detector_size(self):
        """Number of elements currently in the detector dimensions"""
        return shape_to_size(self.current_detector_shape)

    @property
    def current_scan_size(self):
        """Number of elements currently in the shape dimensions"""
        return self.npoints

    @property
    def current_size(self):
        """Number of elements currently in the dataset"""
        return self.current_scan_size * self.current_detector_size

    @property
    def itemsize(self):
        """dtype size in bytes"""
        return self._itemsize

    @property
    def current_bytes(self):
        return self.current_size * self.itemsize

    @property
    def maxshape(self):
        # All dimensions are variable:
        # - Currently detector_shape does not contain
        #   -1 to indicate variable length so assume
        #   any dimension can have a variable length.
        # - Detector may publish more points than
        #   expected.
        return self._maxshape

    def add_external(self, newdata, file_format=None):
        """Add data as external references.

        :param list((str, int)) newdata: uri and index within the uri
        :param str file_format: if not specified, uris will be saved as strings
        """
        if self.is_internal:
            msg = f"{self} already has internal data"
            raise RuntimeError(msg)
        if file_format == "hdf5":
            if self._external_raw or self._external_names:
                msg = f"{self} cannot mix HDF5 with other formats"
                raise RuntimeError(msg)
            self._external_datasets += newdata
            self.npoints = len(self._external_datasets)
        elif file_format in self._external_raw_formats:
            if self._external_datasets or self._external_names:
                msg = f"{self} cannot mix {repr(file_format)} with other formats"
                raise RuntimeError(msg)
            self._external_raw += newdata
            self.npoints = len(self._external_raw)
        else:
            if self._external_datasets or self._external_raw:
                msg = f"{self} cannot mix file formats"
                raise RuntimeError(msg)
            self._external_names += newdata
            self.npoints = len(self._external_names)

    def add(self, newdata):
        """Add data to dataset (copy)

        :param h5py.Dataset dset: shape = scan_shape + detector_shape
        :param sequence newdata: shape = (nnew, ) + detector_shape
        """
        if self.is_external:
            msg = f"{self} already has external data"
            raise RuntimeError(msg)
        super().add(newdata)

    add_internal = add

    def _insert_data(self, dset, newdata: Sequence[numpy.ndarray]):
        """Add data to the internal buffer

        :param h5py.Dataset dset:
        :param sequence newdata: shape = (npoints, ) + detector_shape
        :returns int: number of added points
        """
        nnew = len(newdata)
        points_buffer = self._internal_buffer
        nalready_buffered = len(points_buffer)
        points_buffer.extend(newdata)
        nbuffered = nalready_buffered + nnew
        nalready_saved = self.npoints - nalready_buffered
        if self._save_interal_time is None:
            self._save_interal_time = time.time()

        # Save points aligned with the HDF5 chunks
        nchunk = self._npoints_h5chunk - (nalready_saved % self._npoints_h5chunk)
        nsave = (nbuffered // nchunk) * nchunk
        # Save non-aligned when the data arrives too slow.
        if not nsave:
            if (time.time() - self._save_interal_time) > self._save_interal_dtmax:
                nsave = nbuffered

        if nsave:
            newdata = self._merge_ragged_sequence(points_buffer[:nsave])
            self._internal_buffer = points_buffer[nsave:]
            info = self._insert_data_info(
                dset.shape, nalready_saved, nsave, newdata[0].shape
            )
            self.current_scan_save_shape = info["new_scanshape"]
            self.current_detector_shape = info["new_detshape"]
            self._save_internal_data(dset, newdata, info)
            self._save_interal_time = time.time()
        return nnew

    def _merge_ragged_sequence(self, sequence):
        # Return the sequence when not ragged
        if self.detector_ndim != 1:
            return sequence
        shape0 = sequence[0].shape
        if all(e.shape == shape0 for e in sequence):
            return sequence

        # Regular nD array
        nmax = max(e.shape[0] if e.ndim else 1 for e in sequence)
        shape = (len(sequence), nmax)
        arr = numpy.full(shape, self.fillvalue, dtype=self.dtype)
        for src, dest in zip(sequence, arr):
            dest[: len(src)] = src
        return arr

    def flush(self):
        """Flush any buffered data"""
        # Flush external data (VDS or raw external datasets)
        super().flush()

        # Flush internal data
        if self.is_external or not self._internal_buffer:
            return
        self._npoints_h5chunk = 1
        self.add_internal([])

    def _insert_data_info(self, old_shape, nold_points, nnew_points, newdata_detshape):
        """Add data to the internal buffer

        :param tuple old_shape:
        :param int nnew_points:
        :param tuple newdata_detshape:
        :param sequence newdata: shape = (npoints, ) + detector_shape
        :returns dict:
        """
        scan_ndim = self.scan_save_ndim
        det_ndim = self.detector_ndim
        csaveorder = self.csaveorder

        if csaveorder:
            scan_slice = slice(None, scan_ndim)
            det_slice = slice(scan_ndim, None)
        else:
            det_slice = slice(None, det_ndim)
            scan_slice = slice(det_ndim, None)

        old_scanshape = old_shape[scan_slice]
        old_detshape = old_shape[det_slice]
        icurrent = nold_points
        inext = icurrent + nnew_points
        new_detshape = tuple(max(a, b) for a, b in zip(old_detshape, newdata_detshape))

        if scan_ndim == 0:
            save_coord = None
            if inext == 1:
                new_scanshape = tuple()
            else:
                new_scanshape = (inext,)
        elif scan_ndim == 1:
            save_coord = None
            new_scanshape = (max(old_shape[0], inext),)
        else:
            save_coord, new_scanshape = self._save_shape_mindex(
                icurrent, inext, old_scanshape
            )

        if csaveorder:
            new_shape = new_scanshape + new_detshape
        else:
            new_shape = new_detshape + new_scanshape

        info = {
            "old_shape": old_shape,
            "old_scanshape": old_scanshape,
            "old_detshape": old_detshape,
            "new_shape": new_shape,
            "new_scanshape": new_scanshape,
            "new_detshape": new_detshape,
            "newdata_detshape": newdata_detshape,
            "scan_ndim": scan_ndim,
            "det_ndim": det_ndim,
            "scan_slice": scan_slice,
            "det_slice": det_slice,
            "csaveorder": csaveorder,
            "icurrent": icurrent,
            "inext": inext,
            "nnew_points": nnew_points,
            "save_coord": save_coord,
        }
        return info

    def _save_shape_mindex(self, start, stop, shape):
        """Coordinates in shape (which may need to be expended along the
        slow dimension) that correspond to flat indices `range(start, stop)`.

        :param int start:
        :param int stop:
        :param tuple shape: shape to be filled (with self.saveorder)
        :returns iterable, shape: coordinates and (expanded) shape
        """
        indices = list(range(start, stop))
        while True:
            try:
                save_coord = self.saveorder.unravel(indices, shape)
                break
            except ValueError:
                # Increase the variable dimension or
                # the slow dimension if fixed-length scan
                shape = list(shape)
                try:
                    vdim = shape.index(0)
                except ValueError:
                    if self.csaveorder:
                        vdim = 0
                    else:
                        vdim = -1
                shape[vdim] += 1
                shape = tuple(shape)
        return save_coord, shape

    def _save_internal_data(self, dset, newdata, info):
        """Add data to the HDF5 dataset

        :param h5py.Dataset dset:
        :param sequence newdata:
        :returns int: number of added points
        """
        # Extend HDF5 dataset
        if info["old_shape"] != info["new_shape"]:
            dset.resize(info["new_shape"])

        # Insert new data in HDF5 dataset
        if info["scan_ndim"] == 0:
            dset[()] = newdata
        else:
            if info["csaveorder"]:
                idx = [None] * info["scan_ndim"] + [
                    slice(0, n) for n in info["newdata_detshape"]
                ]
            else:
                idx = [slice(0, n) for n in info["newdata_detshape"]] + [None] * info[
                    "scan_ndim"
                ]
            if info["scan_ndim"] == 1:
                # all at once
                if info["csaveorder"]:
                    idx[0] = slice(info["icurrent"], info["inext"])
                else:
                    idx[-1] = slice(info["icurrent"], info["inext"])
                    newdata = numpy.asarray(newdata)
                    axes = list(range(1, newdata.ndim)) + [0]
                    newdata = numpy.transpose(newdata, axes)
                try:
                    dset[tuple(idx)] = newdata
                except Exception:
                    if not isinstance(newdata, numpy.ndarray):
                        newdata = numpy.asarray(newdata)
                    self.logger.error(
                        "Failed to store data with shape %s in slice %s of dataset with shape %s",
                        newdata.shape,
                        idx,
                        dset.shape,
                    )
                    raise
            else:
                # point per point (could be done better)
                scan_slice = info["scan_slice"]
                for coordi, newdatai in zip(zip(*info["save_coord"]), newdata):
                    idx[scan_slice] = coordi
                    dset[tuple(idx)] = newdatai

    @property
    def current_scan_save_shape(self):
        """This refers to the currently know scan shape, not the current
        shape in terms of arrived data.
        """
        if self.scan_save_ndim == 1:
            return (self.npoints,)
        else:
            return self._current_scan_save_shape

    @current_scan_save_shape.setter
    def current_scan_save_shape(self, value):
        self._current_scan_save_shape = value

    @property
    def compression(self):
        return self._h5_dset_config["compression"]

    @property
    def fillvalue(self):
        """Value reader gets for uninitialized elements"""
        if self.dtype in (str, bytes):
            return ""
        fillvalue = numpy.nan
        try:
            numpy.array(fillvalue, dtype=self.dtype)
        except ValueError:
            fillvalue = 0
        return fillvalue

    @property
    def external_source_args(self):
        nframes = self.external_images_per_file
        if nframes is None:
            # self.logger.warning("Number of frames per external file is not specified")
            return {}
        else:
            shape = (nframes,) + self.detector_shape
            return {"shape": shape, "dtype": self.dtype}

    @property
    def _dset_value(self):
        value = {"fillvalue": self.fillvalue, "dtype": self.dtype}
        if self._external_datasets:
            files, nuris, fill_generator = self._external_dataset_uris()
            if not nuris:
                self.logger.warning("No data to create a virtual dataset")
                return None
            value["data"] = files
            value["order"] = self.saveorder
            value["fill_generator"] = fill_generator
            value["virtual_source_args"] = self.external_source_args
            value["axis"] = 0
            value["newaxis"] = True
            value["maxshape"] = self.maxshape
            if nexus.HASVIRTUAL:
                value["shape"] = self.current_h5shape
                self.logger.info(
                    "merge %d URIs from %d files as a virtual dataset",
                    nuris,
                    len(files),
                )
            else:
                value.update(self._h5_dset_config)
                self.logger.info(
                    "merge %d URIs from %d files as a copy (VDS not supported)",
                    nuris,
                    len(files),
                )
        elif self._external_raw:
            nuris = self._get_external_raw(value)
            if not nuris:
                self.logger.warning("No data to create an external dataset")
                return None
            # Same number of external files as nframes
            nframes = shape_to_size(self.current_scan_save_shape)
            ext = os.path.splitext(self._external_raw[0][0])[-1]
            filename = self._dummy_filename(ext)
            nskip = h5_external.resize(value, nframes, filename, value["fillvalue"])
            if nskip > 0:
                self.logger.warning("Skip %d files", nskip)
            elif nskip < 0:
                self.logger.warning("Missing %d files", nskip)
            # Finalize arguments
            fillorder = self.saveorder
            addorder = Order()
            h5_external.finalize(
                value,
                shape=self.current_scan_save_h5shape,
                addorder=addorder,
                fillorder=fillorder,
            )
            # REMARK: no chunking or reshaping
            #         links are absolute paths
            value["shape"] = self.current_h5shape
            value["chunks"] = None
            nuris = len(value.get("external", []))
            self.logger.info("merge %d URIs as an external (non-HDF5) dataset", nuris)
        elif self._external_names:
            value = []
            dirname = os.path.dirname(self.filename)
            for filename, ind in self._external_names:
                filename = os.path.relpath(filename, dirname)
                value.append(f"{filename}::{ind}")
            self.logger.info("merge %d URIs as a list of strings", len(value))
        else:
            value["shape"] = self.current_h5shape
            value["maxshape"] = self.maxshape
            value.update(self._h5_dset_config)
            self.logger.debug("create as internal data (copy)")
        return value

    def _external_dataset_uris(self):
        """
        :returns list, num, generator:
        """
        # Assume dataset shapes are (nframes, detdim0, detdim1, ...)
        extorder = Order()
        cextorder = extorder.order
        csaveorder = self.csaveorder

        # Map uri to list of (idxin(int or slice), idxout(tuple(int or slice))) tuples
        uridict = OrderedDict()
        uris = self._get_external_datasets()
        if self.current_scan_save_shape:
            coordout = range(len(uris))
            coordout = self.saveorder.unravel(coordout, self.current_scan_save_shape)

            collapse_slice = self.csaveorder

            for (uri, idxin), idxout in zip(uris, zip(*coordout)):
                if uri in uridict:
                    if collapse_slice:
                        self._vdsidx_append(uridict[uri], idxin, idxout)
                    else:
                        uridict[uri].append((idxin, idxout))
                else:
                    uridict[uri] = [(idxin, idxout)]
        else:
            uri, idxin = uris[0]
            uridict[uri] = [(idxin, tuple())]

        # Generates uri's with associated in/out index generator
        def fill_generator():
            for uri, lst in uridict.items():

                def index_generator():
                    for idxin, idxout in lst:
                        idxin = self._add_detidx(idxin, cextorder)
                        idxout = self._add_detidx(idxout, csaveorder)
                        yield idxin, idxout

                yield uri, index_generator

        files = list(uridict.keys())
        nuris = sum(len(v) for v in uridict.values())
        return files, nuris, fill_generator

    @classmethod
    def _vdsidx_append(cls, lst, idxin, idxout):
        """Append VDS input/output index or modify the last one

        :param list(2-tuple) lst:
        :param int idxin:
        :param tuple(int) idxout:
        """
        idxin_prev, idxout_prev = lst[-1]

        # Input index follows the previous one?
        inext = cls._vdsidx_slice_next(idxin_prev, idxin)
        if not inext:
            lst.append((idxin, idxout))
            return

        # Output index follows the previous one
        onext = [
            cls._vdsidx_slice_next(ilast, i) for ilast, i in zip(idxout_prev, idxout)
        ]
        if sum(onext) != 1:
            lst.append((idxin, idxout))
            return
        iout = onext.index(True)

        # Output index the same except for the one that incremented
        eout = [ilast == i for ilast, i in zip(idxout_prev, idxout)]
        eout.pop(iout)
        if not all(eout):
            lst.append((idxin, idxout))
            return

        # Extent output index slice
        idxout_prev = idxout_prev[iout]
        end = idxout[iout]
        if isinstance(idxout_prev, slice):
            start = idxout_prev.start
            step = idxout_prev.step
        else:
            start = idxout_prev
            step = end - start
        idxout = list(idxout)
        idxout[iout] = slice(start, end + step, step)
        idxout = tuple(idxout)

        # Extent input index slice
        if isinstance(idxin_prev, slice):
            a = idxin_prev.start
            b = idxin
            s = idxin_prev.step
        else:
            a = idxin_prev
            b = idxin
            s = b - a
        b += s
        idxin = slice(a, b, s)

        # Replace previous index with extended slices
        lst[-1] = (idxin, idxout)

    @staticmethod
    def _vdsidx_slice_next(idx_prev, idx):
        """Index follows the previous index"""
        step = 1
        if isinstance(idx_prev, slice):
            if idx_prev.step is not None:
                step = idx_prev.step
            idx_prev = next(reversed(range(idx_prev.start, idx_prev.stop, step)))
        # Negative steps not supported by h5py VDS
        return idx_prev + step == idx

    def _add_detidx(self, idx, corder):
        """Add detector dimensions to index"""
        if not isinstance(idx, tuple):
            idx = (idx,)
        if self.detector_ndim:
            if corder:
                return idx + (Ellipsis,)
            else:
                return (Ellipsis,) + idx
        else:
            return idx

    def _get_external_datasets(self):
        if not self.external_uri_from_file:
            return self._external_datasets
        self.logger.info("Retrieving HDF5 URI's ...")
        uris = list(zip(*self._external_datasets))[0]
        filenames = set(nexus.splitUri(uri)[0] for uri in uris)
        uridict = {}
        mon = FileSizeMonitor()
        for filename in sorted(filenames):
            mon.filename = filename
            while True:
                try:
                    uri = nexus.getDefaultUri(filename, locking=True)
                except (RuntimeError, OSError):
                    if not mon.is_growing():
                        raise
                else:
                    if uri:
                        self.logger.debug("Got URI from file %s", filename)
                        uridict[filename] = uri
                        break
                    else:
                        if not mon.is_growing():
                            raise RuntimeError(f"Cannot get URI from file {filename}")
                time.sleep(0.1)
        return [
            (uridict[nexus.splitUri(uri)[0]], i) for uri, i in self._external_datasets
        ]

    def _get_external_raw(self, createkwargs):
        self.logger.info("Retrieving external URI's ...")
        mon = FileSizeMonitor()
        for filename, i in self._external_raw:
            mon.filename = filename
            while True:
                try:
                    h5_external.add_arguments(
                        [(filename, i)], createkwargs=createkwargs
                    )
                except OSError:
                    if not mon.is_growing():
                        raise
                else:
                    break
                time.sleep(0.1)
        return len(createkwargs.get("external", []))

    @property
    def is_external(self):
        """ "External" means a virtual dataset or a raw external dataset,
        for example links to EDF files.
        """
        return bool(
            self._external_datasets or self._external_raw or self._external_names
        )

    @property
    def is_internal(self):
        """ "Internal" mean a normal HDF5 dataset"""
        return self.exists and not self.is_external

    @property
    def interpretation(self):
        return nexus.nxDatasetInterpretation(
            self.scan_ndim, self.detector_ndim, self.ndim
        )

    @property
    def _dset_attrs(self):
        """HDF5 dataset attributes"""
        attrs = self.device.get("data_info", {})
        interpretation = self.interpretation
        if interpretation:
            attrs["interpretation"] = interpretation
        attrs = {k: v for k, v in attrs.items() if v is not None}
        return attrs

    def _create(self, nxroot):
        """Create the dataset"""
        parent = nxroot[self.parent]
        value = self._dset_value
        if value:
            attrs = self._dset_attrs
            dset = nexus.nxCreateDataSet(parent, self.name, value, attrs)
            if self.alias:
                nexus.createLink(parent, self.alias, dset)
            self._exists = True

    def _create_parent(self, nxroot):
        raise RuntimeError("Parent must exist already")

    @property
    def _progress_log_string(self):
        return f" ({format_bytes(self.current_bytes)})"

    def reshape(self, scan_save_shape, detector_shape=None):
        """Reshape dataset (must exist when internal, should not exist when external)

        :param tuple or None scan_save_shape:
        :param tuple or None detector_shape:
        """
        if scan_save_shape is None:
            scan_save_shape = self.current_scan_save_shape
        elif len(scan_save_shape) != self.scan_save_ndim:
            raise ValueError("Number of scan dimensions should not change")
        if detector_shape is None:
            detector_shape = self.current_detector_shape
        elif len(detector_shape) != self.detector_ndim:
            raise ValueError("Number of detector dimensions should not change")
        if self.is_external:
            reshaped = self._reshape_external(scan_save_shape, detector_shape)
        else:
            reshaped = self._reshape_internal(scan_save_shape, detector_shape)
        if reshaped:
            self.current_scan_save_shape = scan_save_shape
            self.current_detector_shape = detector_shape
            self.npoints = shape_to_size(scan_save_shape)

    def _reshape_internal(self, scan_save_shape, detector_shape):
        """Reshape HDF5 dataset if it exists

        :param tuple or None scan_save_shape:
        :param tuple or None detector_shape:
        :return bool:
        """
        with self.open(create=False) as dset:
            if dset is None:
                self.logger.warning("Cannot reshape internal dataset before creation")
                return False
            shape = dset.shape
            if self.csaveorder:
                new_shape = scan_save_shape + detector_shape
            else:
                new_shape = detector_shape + scan_save_shape
            if dset.shape != new_shape:
                self.logger.info("reshape from %s to %s", shape, new_shape)
                try:
                    dset.resize(new_shape)
                except TypeError:
                    self.logger.exception("Dataset cannot be reshaped")
                else:
                    return True
        return False

    def _reshape_external(self, scan_save_shape, detector_shape):
        """Reshape HDF5 dataset if it exists

        :param tuple or None scan_save_shape:
        :param tuple or None detector_shape:
        :return bool:
        """
        if self.exists:
            self.logger.warning("Cannot reshape external dataset after creation")
            return False
        npoints = shape_to_size(scan_save_shape)
        if self._external_datasets:
            lst = self._external_datasets
        elif self._external_names:
            lst = self._external_names
        else:
            lst = self._external_raw
        nlst = len(lst)
        if npoints != nlst:
            if npoints < nlst:
                nremove = nlst - npoints
                self.logger.info("remove %d points", nremove)
                lst = lst[:npoints]
            else:
                nadd = npoints - nlst
                self.logger.info("add %d dummy points", nadd)
                lst += self._dummy_uris(nadd)
            if self._external_datasets:
                self._external_datasets = lst
            elif self._external_names:
                self._external_names = lst
            else:
                self._external_raw = lst
            return True
        return False

    def _dummy_uris(self, npoints):
        """URIs of dummy data (create when missing).

        :returns lst(str, int): uri, index
        """
        hdf5 = bool(self._external_datasets)
        if hdf5:
            ext = ".h5"
        else:
            ext = ".edf"
        filename = self._dummy_filename(ext)
        if not os.path.isfile(filename):
            mkdir(os.path.dirname(filename))
            fillvalue = self.fillvalue
            dtype = self.dtype
            if hdf5:
                shape = (1,) + self.current_detector_shape
                with nexus.File(filename, mode="w") as f:
                    dset = f.create_dataset(
                        "data", shape=shape, dtype=dtype, fillvalue=fillvalue
                    )
                    nexus.markDefault(dset)
            else:
                shape = self.current_detector_shape
                if not shape:
                    shape = 1, 1
                elif len(shape) == 1:
                    shape = 1, shape[0]
                data = numpy.full(shape, fillvalue, dtype=dtype)
                edf = EdfImage(data=data, header=None)
                edf.write(filename)
        return [(filename, 0) for _ in range(npoints)]

    def _dummy_filename(self, ext):
        """
        :param str ext: .edf, .h5, ...
        :returns str:
        """
        dirname = os.path.dirname(self.filename)
        name = "_".join(map(str, self.current_detector_shape)) + "_" + self.dtype_name
        return os.path.join(dirname, "dummy", "dummy_" + name + ext)

    def add_metadata(self, treedict, parent=False, create=False, **kwargs):
        """Add datasets/attributes (typically used for metadata)

        :param dict treedict:
        :param bool parent:
        :param bool create: create destination when it does not exist
        :param kwargs: see `dicttonx`
        """
        if parent:
            ctx = self.open_parent
        else:
            ctx = self.open
        with ctx(create=create) as destination:
            if destination is None:
                self.logger.error("Cannot add metadata before creation")
            else:
                dictdump.dicttonx(treedict, destination, **kwargs)



"""Activate Bliss session utilities
"""

from functools import wraps


def with_scan_saving(func):
    """Pass the current session's SCAN_SAVING instance as a named argument

    :param callable func:
    :returns callable:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        scan_saving = kwargs.get("scan_saving")
        if scan_saving is None:
            from bliss import current_session  # avoid bliss patching on import

            if current_session:
                kwargs["scan_saving"] = current_session.scan_saving
            else:
                raise RuntimeError("No activate Bliss session")
        return func(*args, **kwargs)

    return wrapper


@with_scan_saving
def scan_saving_get(attr, default=None, scan_saving=None):
    """Get attribute from the session's scan saving object

    :param str attr:
    :param default:
    :param bliss.scanning.scan.ScanSaving scan_saving:
    :returns str:
    """
    return getattr(scan_saving, attr, default)


@with_scan_saving
def dataset_get(attr, default=None, scan_saving=None):
    """Get attribute from the session's dataset object

    :param str attr:
    :param default:
    :param bliss.scanning.scan.ScanSaving scan_saving:
    :returns str:
    """
    try:
        return scan_saving.dataset[attr]
    except (AttributeError, KeyError):
        return default



"""Compile device information before and after Redis publication
"""

from collections import OrderedDict, namedtuple

DeviceTypeDefinition = namedtuple(
    "DeviceTypeDefinition",
    ["namemap", "typemap", "unitmap"],
)

DEFAULT_TYPE_DEFINITION = DeviceTypeDefinition(
    namemap=dict(), typemap=dict(), unitmap=dict()
)

TYPE_DEFINITIONS = {
    "mca": DeviceTypeDefinition(
        namemap={  # See https://gitlab.esrf.fr/-/snippets/249
            "icr": "trigger_count_rate",  # triggers/trigger_live_time
            "ocr": "event_count_rate",  # events/elapsed_time
            "triggers": "triggers",
            "events": "events",
            "deadtime": "fractional_dead_time",
            "trigger_livetime": "trigger_live_time",
            "energy_livetime": "live_time",
            "realtime": "elapsed_time",
            "det": "spectrum",
        },
        typemap={
            "spectrum": "principal",
            "icr": "icr",
            "ocr": "ocr",
            "triggers": "triggers",
            "events": "events",
            "deadtime": "deadtime",
            "trigger_livetime": "trigger_livetime",
            "energy_livetime": "livetime",
            "realtime": "realtime",
        },
        unitmap={
            "icr": "hertz",
            "orc": "hertz",
            "livetime": "s",
            "trigger_livetime": "s",
            "realtime": "s",
        },
    ),
    "mosca": DeviceTypeDefinition(
        namemap={  # See https://gitlab.esrf.fr/-/snippets/249
            "icr": "trigger_count_rate",  # triggers/trigger_live_time
            "ocr": "event_count_rate",  # events/elapsed_time
            "triggers": "triggers",
            "events": "events",
            "deadtime": "fractional_dead_time",
            "trigger_livetime": "trigger_live_time",
            "energy_livetime": "live_time",
            "realtime": "elapsed_time",
            "det": "spectrum",
        },
        typemap={
            "spectrum": "principal",
            "icr": "icr",
            "ocr": "ocr",
            "triggers": "triggers",
            "events": "events",
            "deadtime": "deadtime",
            "trigger_livetime": "trigger_livetime",
            "energy_livetime": "livetime",
            "realtime": "realtime",
        },
        unitmap={
            "icr": "hertz",
            "orc": "hertz",
            "livetime": "s",
            "trigger_livetime": "s",
            "realtime": "s",
        },
    ),
    "lima": DeviceTypeDefinition(
        namemap=dict(),
        typemap={"image": "principal", "sum": "principal"},
        unitmap=dict(),
    ),
    "mythen": DeviceTypeDefinition(
        namemap=dict(), typemap={"spectrum": "principal"}, unitmap=dict()
    ),
    "samplingcounter": DeviceTypeDefinition(
        namemap=dict(), typemap={"signal": "principal"}, unitmap=dict()
    ),
}


def shortnamemap(names, separator=":"):
    """Map full Redis names to short (but still unique) names

    :param list(str) names:
    :param str separator:
    :returns dict:
    """
    if not names:
        return {}
    names = set(names)
    parts = [name.split(separator) for name in names]
    nparts = max(map(len, parts))
    parts = [([""] * (nparts - len(lst))) + lst for lst in parts]
    ret = {}
    for i in reversed(range(-nparts, 0)):
        joinednames = [separator.join(s for s in lst[i:] if s) for lst in parts]
        newnames = joinednames + list(ret.values())
        selection = [
            (idx, (separator.join(s for s in lst if s), name))
            for idx, (name, lst) in enumerate(zip(joinednames, parts))
            if newnames.count(name) == 1
        ]
        if selection:
            idx, tuples = list(zip(*selection))
            ret.update(tuples)
            parts = [lst for j, lst in enumerate(parts) if j not in idx]
    return ret


def fill_device(fullname, device, device_info=None, data_info=None):
    """Add missing keys with default values

        device_type: type for the writer (not saved), e.g. positioner, mca, lima
        device_name: HDF5 group name (measurement or positioners when missing)
        device_info: HDF5 group datasets
        data_type: maximal one device dataset has type "principal"
        data_name: HDF5 dataset name
        data_alias: HDF5 dataset link name
        data_info: HDF5 dataset attributes
        unique_name: Unique name for HDF5 links
        master_index: >=0 axis order used for plotting
        single: only data of this device
                a single's data_name is either "data" (detector) or "value" (positioner)
        dependencies: fullnames

    :param str fulname:
    :param dict device:
    :param dict device_info:
    :param dict data_info:
    """
    if device_info is None:
        device_info = {}
    if data_info is None:
        data_info = {}
    device["device_type"] = device.get("device_type", "")
    device["device_name"] = device.get("device_name", fullname)
    device["device_info"] = device.get("device_info", device_info)
    device["data_type"] = device.get("data_type", "principal")
    device["data_name"] = device.get("data_name", "data")
    device["data_alias"] = device.get("data_alias", None)
    device["data_info"] = device.get("data_info", data_info)
    device["unique_name"] = device.get("unique_name", fullname)
    device["master_index"] = -1
    device["single"] = None
    device["dependencies"] = {}
    device["metadata_keys"] = {}


def update_device(devices, fullname, device_info=None, data_info=None):
    """Add missing device and/or keys

    :param dict devices:
    :param str fullname:
    :param dict device_info:
    :param dict data_info:
    """
    devices[fullname] = device = devices.get(fullname, {})
    fill_device(fullname, device, device_info=device_info, data_info=data_info)
    return device


def parse_devices(devices, short_names=True, multivalue_positioners=False):
    """Determine names and types based on device name and type

    :param dict devices:
    :param bool short_names:
    :param bool multivalue_positioners:
    """
    namemap = shortnamemap(list(devices.keys()))
    for fullname, device in devices.items():
        device["device_name"] = namemap.get(fullname, fullname)
        type_definition = TYPE_DEFINITIONS.get(
            device["device_type"], DEFAULT_TYPE_DEFINITION
        )
        if device["device_type"] == "mca":
            # 'xmap1:xxxxxx_det1'
            # 'xmap1:roi1'
            #
            #   xxxxxx: spectrum, icr, ocr, triggers, events,
            #           deadtime, trigger_livetime, energy_livetime,
            #           realtime, roi1, roi2, ...
            #
            #   device_name = 'xmap1:det1'
            #   data_type = mcatypemap('xxxxxx')
            #   data_name = mcanamemap('xxxxxx')
            parts = fullname.split(":")
            lastparts = parts[-1].split("_")
            if len(lastparts) == 1:
                mcachannel = "sum"
                datatype = lastparts[-1]  # roi1
            else:
                mcachannel = lastparts[-1]
                datatype = "_".join(lastparts[:-1])  # xxxxxx

            parts = parts[:-1] + [mcachannel]
            device["device_name"] = ":".join(parts)
            device["data_name"] = type_definition.namemap.get(datatype, datatype)
            if datatype.startswith("roi"):
                datatype = "roi"
            device["data_type"] = type_definition.typemap.get(datatype, datatype)
            device["data_info"]["units"] = type_definition.unitmap.get(
                device["data_type"], None
            )
        elif device["device_type"] == "mosca":
            # 'falconx:spectrum:det01'
            #
            # 'falconx:stat:xxxx_det01'
            #  with xxxx in: [ icr, ocr, triggers, events, deadtime, trigger_livetime, energy_livetime, realtime ]
            #
            # 'falconx:roi:xxx_det01' or 'falconx:roi:xxx_sum_all' or 'falconx:roi:xxx_sum_yy_zz'
            #
            parts = fullname.split(":")
            lastparts = parts[-1].split("_")
            if parts[-2] == "spectrum":
                mcachannel = f"det{int(lastparts[-1][-2:])}"
                datatype = "spectrum"
            elif parts[-2] == "stat":
                mcachannel = f"det{int(lastparts[-1][-2:])}"
                datatype = "_".join(lastparts[:-1])  # xxxx
            elif parts[-2] == "roi":
                channel = device.get("channel")
                if isinstance(channel, tuple):
                    cfrom, cto = channel
                    mcachannel = f"sum_{cfrom}_{cto}"
                    datatype = "_".join(lastparts[:-3])  # xxx
                elif channel == -1:
                    mcachannel = "sum_all"
                    datatype = "_".join(lastparts[:-2])  # xxx
                else:
                    mcachannel = f"det{channel}"
                    datatype = "_".join(lastparts[:-1])  # xxx

            device["data_name"] = type_definition.namemap.get(datatype, datatype)
            datatype = "roi" if parts[-2] == "roi" else datatype
            device["data_type"] = type_definition.typemap.get(datatype, datatype)
            parts = parts[:-2] + [mcachannel]
            device["device_name"] = ":".join(parts)
            device["data_info"]["units"] = type_definition.unitmap.get(
                device["data_type"], None
            )
        elif device["device_type"] == "mythen":
            # 'mythen1:spectrum'
            # 'mythen1:roi1'
            parts = fullname.split(":")
            device_name = parts[0]
            datatype = ":".join(parts[1:])
            device["device_name"] = device_name
            device["data_type"] = type_definition.typemap.get(datatype, datatype)
            device["data_name"] = type_definition.namemap.get(datatype, datatype)
        elif device["device_type"] == "lima":
            # 'frelon1:image'
            # 'frelon1:roi_counters:roi1_min'
            # 'frelon1:roi_profiles:roi2'
            # 'frelon1:bpm:fwhm_x'
            # 'frelon1:roi_collection:roi_collection_counter_data'
            parts = fullname.split(":")
            if len(parts) == 3:
                device["dependencies"] = {parts[0] + ":image": "image"}
                if parts[1] == "roi_counters":
                    roi_name, _, processing_name = parts[-1].rpartition("_")
                    device_name = ":".join([parts[0], roi_name])
                    datatype = processing_name
                    device["metadata_keys"] = {roi_name: "selection"}
                elif parts[1] == "roi_profiles":
                    device_name = ":".join([parts[0], parts[-1]])
                    datatype = "sum"
                    device["metadata_keys"] = {parts[-1]: "selection"}
                elif parts[1] == "roi_collection":
                    device_name = ":".join([parts[0], parts[1]])
                    datatype = "sum"
                    device["metadata_keys"] = {"roi_collection_counter": "selection"}
                else:
                    device_name = ":".join(parts[:2])
                    datatype = parts[2]
            else:
                device_name = parts[0]
                datatype = ":".join(parts[1:])
            device["device_name"] = device_name
            device["data_type"] = type_definition.typemap.get(datatype, datatype)
            device["data_name"] = type_definition.namemap.get(datatype, datatype)
        elif device["device_type"] == "autofilter":
            # autofilter1:xxx
            #  device_name = 'autofilter1'
            #   data_type = typemap('xxxxxx')
            #   data_name = namemap('xxxxxx')
            device["device_name"], datatype = fullname.split(":")
            device["data_type"] = type_definition.typemap.get(datatype, datatype)
            device["data_name"] = type_definition.namemap.get(datatype, datatype)
        elif device["device_type"] == "samplingcounter":
            if device["data_type"] == "signal":
                datatype = "signal"
            else:
                # 'simdiodeSAMPLES_xxxxx'
                #   device_name = 'simdiodeSAMPLES'
                #   data_type = typemap('xxxxxx')
                #   data_name = namemap('xxxxxx')
                parts = device["device_name"].split("_")
                datatype = parts[-1]  # xxxxxx
                parts = ["_".join(parts[:-1])]
                device["device_name"] = "_".join(parts)
            device["data_type"] = type_definition.typemap.get(datatype, datatype)
            device["data_name"] = type_definition.namemap.get(datatype, datatype)
        elif device["device_type"] == "positionergroup":
            parts = fullname.split(":")
            if device["data_type"] != "principal":
                device["data_type"] = parts[-1]
            if multivalue_positioners:
                device["device_name"] = ":".join(parts[:-1])
                if device["data_type"] == "principal":
                    device["data_name"] = "value"
                else:
                    device["data_name"] = device["data_type"]
            else:
                device["data_name"] = "value"
                device["single"] = True
        elif device["device_type"] == "positioner":
            device["data_name"] = "value"
            device["data_type"] = "principal"
        else:
            device["data_name"] = "data"
            device["data_type"] = "principal"
        if device["single"] is None:
            device["single"] = device["data_type"] == "principal"
        if device["data_type"] == "principal":
            if device["device_type"] in ("positionergroup", "positioner"):
                device["data_alias"] = "value"
            else:
                device["data_alias"] = "data"
        if device["single"]:
            device["unique_name"] = device["device_name"]
        else:
            device["unique_name"] = device["device_name"] + ":" + device["data_name"]


Device = namedtuple(
    "Device",
    ["subscan", "fullname", "info", "is_positioner", "in_positioner_group", "index"],
)


def iter_devices(scan_info, scan_ndim=1):
    """
    :param dict scan_info:
    :param int scan_ndim:
    :yields Device:
    """
    for subscan_name, subscan_info in scan_info["acquisition_chain"].items():
        ntriggers = 0
        for group_name in subscan_info["devices"]:
            group_info = scan_info["devices"][group_name]
            is_trigger = len(group_info.get("triggered_devices", [])) > 0
            if is_trigger:
                ntriggers += 1
            is_first_trigger = is_trigger and ntriggers == 1
            group_size = len(group_info["channels"])
            for i, fullname in enumerate(group_info["channels"]):
                info = scan_info["channels"].get(fullname, dict())
                is_positioner = is_first_trigger and info.get("dim") == 0
                in_positioner_group = (
                    is_positioner
                    and scan_ndim <= 1
                    and group_size > 1
                    and ":" in fullname
                )
                yield Device(
                    subscan=subscan_name,
                    fullname=fullname,
                    info=info,
                    is_positioner=is_positioner,
                    in_positioner_group=in_positioner_group,
                    index=i,
                )


def device_info(
    devices, scan_info, scan_ndim=1, short_names=True, multivalue_positioners=False
):
    """Merge device information from `writer_config_publish.device_info`
    and from the scan info published by the Bliss core library.

    :param dict devices: as provided by `writer_config_publish.device_info`
    :param dict scan_info:
    :param int scan_ndim: number of scan dimensions
    :param bool short_names:
    :param bool multivalue_positioners:
    :returns dict: subscanname:dict(fullname:dict)
                   ordered according to position in
                   acquisition chain
    """
    ret = OrderedDict()
    device_save_keys = {"unit": "units"}
    for cdevice in iter_devices(scan_info, scan_ndim=scan_ndim):
        # Devices are grouped per subscan
        if cdevice.subscan not in ret:
            subscan_devices = ret[cdevice.subscan] = dict()

        # Initial device information from `writer_config_publish.device_info`
        subscan_devices[cdevice.fullname] = devices.get(cdevice.fullname, dict())

        # Add device information
        data_info = {
            kset: cdevice.info.get(kget, None)
            for kget, kset in device_save_keys.items()
        }
        device = update_device(subscan_devices, cdevice.fullname, data_info=data_info)

        # Add extra information for positioners
        if cdevice.is_positioner:
            if cdevice.in_positioner_group:
                device["device_type"] = "positionergroup"
                device["master_index"] = 0
            else:
                device["device_type"] = "positioner"
                device["master_index"] = cdevice.index
            if cdevice.index == 0:
                device["data_type"] = "principal"
            else:
                device["data_type"] = ""

    # Parse/normalize device information
    for devices in ret.values():
        parse_devices(
            devices,
            short_names=short_names,
            multivalue_positioners=multivalue_positioners,
        )
    return ret



import os
from typing import Mapping, Optional, List, Dict


def scan_filename(scan_info: Mapping, raw: bool = False) -> Optional[str]:
    """Filename which contains bliss scans."""
    scan_filename = scan_info.get("filename")
    if not scan_filename or not raw:
        return scan_filename
    writer_options = scan_info.get("writer_options")
    if not writer_options:
        return scan_filename
    if not writer_options.get("separate_scan_files"):
        return scan_filename
    images_path = scan_info.get("images_path")
    if not images_path:
        return scan_filename
    root_dir, scan_basename = os.path.split(scan_filename)
    _, ext = os.path.splitext(scan_basename)
    master_base = "bliss_master" + ext
    subdir = os.path.relpath(images_path, root_dir)
    parts = list()
    for s in subdir.split(os.path.sep):
        if "{" in s:
            break
        if s:
            parts.append(s)
    return os.path.join(root_dir, *parts, master_base)


def scan_filenames(scan_info: Mapping, config: bool = True) -> Dict[str, str]:
    """Names of the files that contain the scan data (raw or as links)"""
    filenames = dict(scan_master_filenames(scan_info, config=config))
    filename = scan_filename(scan_info, raw=False)
    if filename:
        filenames["dataset"] = filename
    filename = scan_filename(scan_info, raw=True)
    if filename:
        filenames["scan"] = filename  # can be the same as dataset
    return filenames


def scan_master_filenames(scan_info: Mapping, config: bool = True) -> Dict[str, str]:
    """Names of the files that contain links to the scan data"""
    if not config:
        return dict()
    info = scan_info.get("nexuswriter", dict())
    return info.get("masterfiles", dict())


def scan_urls(scan_info: Mapping, subscan: int = None, raw: bool = False) -> List[str]:
    """URL of all subscans of a scan"""
    if subscan is None:
        nsubscans = len(scan_info["acquisition_chain"])
        subscans = range(1, nsubscans + 1)
    else:
        subscans = [subscan]
    return list(
        filter(
            None,
            [subscan_url(scan_info, subscan=subscan, raw=raw) for subscan in subscans],
        )
    )


def subscan_url(
    scan_info: Mapping, subscan: int = 1, raw: bool = False
) -> Optional[str]:
    """URL of one subscan of a scan"""
    filename = scan_filename(scan_info, raw=raw)
    if not filename:
        return None
    name = scan_name(scan_info, subscan)
    if not name:
        return None
    return f"{filename}::/{name}"


def scan_name(scan_info: Mapping, subscan: int = 1) -> Optional[str]:
    """Name of the scan in the file tree structure"""
    scan_nb = scan_info["scan_nb"]
    if scan_nb > 0:
        return f"{scan_nb}.{subscan}"



from typing import Tuple, NewType
from numbers import Integral

import numpy
from numpy.typing import DTypeLike

PositiveIntegral = NewType("PositiveIntegral", Integral)  # >= 0
StrictPositiveIntegral = NewType("StrictPositiveIntegral", Integral)  # > 0
ShapeType = Tuple[StrictPositiveIntegral]
SizeType = PositiveIntegral


def dtype_nbytes(dtype: DTypeLike) -> int:
    return numpy.dtype(dtype).itemsize


def shape_to_size(shape: ShapeType) -> int:
    return numpy.prod(shape, dtype=int)


def shape_to_nbytes(shape: ShapeType, dtype: DTypeLike) -> int:
    return shape_to_size(shape) * dtype_nbytes(dtype)


def format_bytes(size: SizeType) -> str:
    power = 1 << 10
    n = 0
    power_labels = {0: "B", 1: "KB", 2: "MB", 3: "GB", 4: "TB"}
    while size >= power and n < 4:
        size /= power
        n += 1
    return "{:.01f}{}".format(size, power_labels[n])



"""Save options for a writer which relies on extra metadata published by Bliss"""

from typing import Dict, Any
from . import base

cli_saveoptions = dict(base.cli_saveoptions)
cli_saveoptions["stackmca"] = {
    "dest": "stack_mcas",
    "action": "store_true",
    "help": "Merged MCA datasets in application definition",
}


def default_saveoptions() -> Dict[str, Any]:
    return base.extract_default_saveoptions(cli_saveoptions)



import pytest

from blisswriter.io.io_utils import rotatefiles


@pytest.mark.parametrize("nmax", [1, 3])
def test_rotating_file(nmax, tmpdir):
    n = nmax + 2
    for i in range(n):
        filename = str(tmpdir.join("logfile.log"))
        rotatefiles(filename, nmax=nmax)
        with open(filename, mode="a") as fp:
            fp.write(str(i))
    for i in range(nmax):
        if i:
            filename = tmpdir.join(f"logfile.{i}.log")
        else:
            filename = tmpdir.join("logfile.log")
        with open(str(filename), mode="r") as fp:
            assert fp.read() == str(n - 1 - i)



from copy import deepcopy
from contextlib import ExitStack
from typing import Dict, Union, Optional, Tuple

import numpy

from blissdata.redis_engine.scan import Scan as RedisScan
from blissdata.lima.client import lima_client_factory

from . import types
from . import typing
from .scan import ScanWriter
from .info import SubScanInfo
from .capture import capture_exceptions
from blisswriter.writer import scan_url_info
from blisswriter.writer.datasets import Hdf5DatasetWriter
from blisswriter.writer.references import Hdf5ScanReferenceWriter


class ScanWriterWithState:
    """Usage

    .. code:: python

        writer = ScanWriterWithState(identifier, name, ...)
        writer.initialize()
        # Optional channel initialization:
        writer.add_channel(channel1)
        writer.add_channel(channel2)
        ...
        writer.add_channel_data(channel1, data)
        writer.add_channel_data(channel2, data)
        ...
        writer.finalize()
    """

    def __init__(self, *args, **kw) -> None:
        self._scan_writer = ScanWriter(*args, **kw)
        self._scan_info = self._scan_writer.scan_info
        self._exit_stack = ExitStack()

        self._channel_to_subscan: Dict[str, SubScanInfo] = dict()
        self._subscan_to_pending_data: Dict[Dict[str, typing.ChannelDataType]] = dict()
        self._lima_clients: Dict[str, tuple(object, bool)] = dict()
        self._lima_proxy = None

    @property
    def scan_writer(self) -> ScanWriter:
        return self._scan_writer

    def initialize(self, scan_info: dict, data_store) -> bool:
        """Call when the scan is prepared"""
        self._data_store = data_store
        self._exit_stack = ExitStack().__enter__()

        self._scan_info.update(deepcopy(scan_info))
        if not self._scan_info.saving_enabled():
            return False

        # Initialize scan writing
        self._scan_info.logger.info(
            "Start writing to '%s' with options %s",
            self._scan_info.filename,
            self._scan_writer.get_save_options(),
        )

        if self._scan_writer._hold_file_open:
            ctx = self._scan_writer.nxroot()
            _ = self._exit_stack.enter_context(ctx)

        # Initialize subscan writing
        with capture_exceptions() as capture:
            for name in self._scan_info._expected_subscans:
                subscan = self._scan_info.add_subscan(name, name)

                adict = self._scan_info["acquisition_chain"][name]
                for device in adict["devices"]:
                    for channel in self._scan_info["devices"][device]["channels"]:
                        self._channel_to_subscan[channel] = subscan

                self._subscan_to_pending_data[subscan.identifier] = dict()

                with capture():
                    self._initialize_subscan(scan_info, subscan)

        return True

    def finalize(self, scan_info: dict, failed: bool) -> None:
        """Call at the end of the scan"""
        try:
            self._scan_info = self._scan_writer.scan_info
            if not self._scan_info.saving_enabled():
                return

            self._scan_info.logger.info("Finalize scan")

            self._scan_info.update(deepcopy(scan_info))

            with capture_exceptions() as capture:
                for _, subscan in self._scan_info.iter_subscans():
                    with capture():
                        self._flush_pending_data(subscan)
                    with capture():
                        self._finalize_subscan(scan_info, failed, subscan)
        finally:
            self._exit_stack.__exit__(None, None, None)
            self._exit_stack = None

    def _initialize_subscan(self, scan_info: dict, subscan: SubScanInfo):
        """One subscan is saved in one NXentry"""
        subscan.logger.info("Initialize subscan")
        with capture_exceptions() as capture:
            with capture():
                subscan.update(deepcopy(scan_info))
            with capture():
                self._scan_writer._update_subscan_status(subscan, "STARTING")
            with capture():
                self._scan_writer._save_positioners(subscan)
            with capture():
                self._scan_writer._create_master_links(subscan)
            with capture():
                self._scan_writer._save_device_metadata(subscan)
            with capture():
                self._scan_writer._save_subscan_user_metadata(subscan)
            with capture():
                self._scan_writer._save_subscan_metadata(subscan)
            with capture():
                if not capture.failed:
                    self._scan_writer._update_subscan_status(subscan, "RUNNING")

    def _finalize_subscan(
        self,
        scan_info: dict,
        failed: bool,
        subscan: SubScanInfo,
    ):
        subscan.logger.info("Finalize subscan")
        with capture_exceptions() as capture:
            with capture():
                subscan.flush()
            with capture():
                subscan.update(deepcopy(scan_info))
            with capture():
                self._scan_writer._save_positioners(subscan)
            with capture():
                self._scan_writer._create_plots(subscan)
            with capture():
                self._scan_writer._save_device_metadata(subscan)
            with capture():
                self._scan_writer._save_subscan_user_metadata(subscan)
            with capture():
                self._scan_writer._save_subscan_notes(subscan)
            with capture():
                self._scan_writer._save_subscan_metadata(subscan)
            with capture():
                self._scan_writer._save_applications(subscan)
            with capture():
                self._scan_writer._save_scan_status(subscan)
            with capture():
                if failed or capture.failed:
                    self._scan_writer._update_subscan_status(subscan, "FAILED")
                else:
                    self._scan_writer._update_subscan_status(subscan, "SUCCEEDED")

    def add_channel(self, channel: types.Channel) -> None:
        """Call when the channel is ready (optional)"""
        subscan = self._channel_to_subscan[channel.name]
        _ = self._get_writer(channel, subscan)

    def _get_writer(
        self,
        channel: types.Channel,
        subscan: SubScanInfo,
        requires_full_channel_info: bool = True,
    ) -> Union[Hdf5DatasetWriter, Hdf5ScanReferenceWriter]:
        if channel.data_type == types.ChannelDataType.SCAN_REFERENCE:
            return self._get_reference_writer(channel, subscan)
        return self._get_dataset_writer(
            channel, subscan, requires_full_channel_info=requires_full_channel_info
        )

    def _get_reference_writer(
        self,
        channel: types.Channel,
        subscan: SubScanInfo,
    ) -> Optional[Hdf5ScanReferenceWriter]:
        writer = subscan.get_reference_writer(channel.name)
        if writer is not None:
            return writer
        return self._scan_writer.add_reference(subscan, channel.name)

    def _get_dataset_writer(
        self,
        channel: types.Channel,
        subscan: SubScanInfo,
        requires_full_channel_info: bool = True,
    ) -> Optional[Hdf5DatasetWriter]:
        # Already initialized?
        writer = subscan.get_dataset_writer(channel.name)
        if writer is not None:
            return writer

        # Can be initialized?
        data_info, complete = self._detector_data_info(channel)
        if requires_full_channel_info and not complete:
            return None

        # Create parent: NXdetector, NXpositioner or measurement
        device = self._scan_writer.scan_info.get_device_info(subscan, channel.name)
        if device is None:
            return None

        with self._scan_writer.device_parent_context(subscan, device) as parent:
            if parent is None:
                return None
            # Save info associated to the device (not this specific dataset)
            if not parent.name.endswith("measurement"):
                self._scan_writer._device_dicttonx_helper(parent, device["device_info"])
            parent = parent.name

        # Everything is ready to create the dataset
        return self._scan_writer.add_dataset(
            subscan, channel.name, parent, device, data_info
        )

    def _detector_data_info(self, channel: types.Channel) -> Tuple[dict, bool]:
        info = dict()
        is_lima = channel.data_type == types.ChannelDataType.LIMA_STATUS
        if is_lima:
            images_per_file = channel.info["lima_info"].get("frame_per_file")
            if images_per_file:
                info["external_images_per_file"] = images_per_file
        else:
            info["external_images_per_file"] = None

        dtype = channel.info["dtype"]
        if dtype is not None:
            info["dtype"] = dtype

        detector_shape = channel.info["shape"]
        if detector_shape is not None:
            # Cannot handle variable dimensions in the detector for lima
            # because we are not allowed to check the actual lima data,
            # we only have references. We can handle variable dimensions
            # for other detectors because we have the data in memory.
            if not is_lima or all(n > 0 for n in detector_shape):
                info["detector_shape"] = detector_shape

        # Fill missing info
        allkeys = {"dtype", "detector_shape", "external_images_per_file"}
        complete = set(info.keys()) == allkeys
        if not complete:
            info.setdefault("detector_shape", tuple())  # 0D detector
            info.setdefault("dtype", float)
            info.setdefault("external_images_per_file", None)
        return info, complete

    def add_channel_data(
        self, channel: types.Channel, data: typing.ChannelDataType
    ) -> bool:
        """Call when the channel has data. Returns `False` when the data is buffered because
        the channel does not have enough information yet."""
        subscan = self._channel_to_subscan[channel.name]
        writer = self._get_writer(channel, subscan, requires_full_channel_info=True)

        pending_data = self._subscan_to_pending_data[subscan.identifier]
        pending = pending_data.get(channel.name)

        # To test pending data:
        # writer = None

        if writer is not None:
            if pending is not None:
                # First flush the pending data
                del pending_data[channel.name]
                self._add_data_to_writer(subscan, channel, writer, pending.get_data())
            self._add_data_to_writer(subscan, channel, writer, data)
            return True

        # The writer could not be instantiated yet. Buffer the data locally.
        if pending is None:
            cls = _CHANNEL_TYPE_TO_PENDING_CLASSES[channel.data_type]
            pending_data[channel.name] = cls(channel, data)
            return False

        pending.add_data(data)
        return False

    def _force_add_channel_data(
        self,
        channel: types.Channel,
        data: typing.ChannelDataType,
    ) -> None:
        """Add data to a writer (force initialization)."""
        subscan = self._channel_to_subscan[channel.name]
        writer = self._get_writer(channel, subscan, requires_full_channel_info=False)
        self._add_data_to_writer(subscan, channel, writer, data)

    def _add_data_to_writer(
        self,
        subscan: SubScanInfo,
        channel: types.Channel,
        writer: Union[Hdf5DatasetWriter, Hdf5ScanReferenceWriter],
        data: typing.ChannelDataType,
    ) -> None:
        """The data types are

        * `numpy.ndarray`: normal data
        * `list`: buffered normal data or ragged data (e.g. diode samples)
        * `dict`: lima status or reference to another scan
        """
        npoints_before = writer.npoints
        try:
            if channel.data_type == types.ChannelDataType.SCAN_REFERENCE:
                for scan_reference in data:
                    scan = self._data_store.load_scan(
                        scan_reference["key"], scan_cls=RedisScan
                    )
                    if scan.info.get("save"):
                        urls = scan_url_info.scan_urls(scan.info)
                        writer.add_references(urls)
            elif channel.data_type == types.ChannelDataType.LIMA_STATUS:
                try:
                    lima_client, use_references = self._lima_clients[channel.name]
                except KeyError:
                    lima_client = lima_client_factory(self._data_store, channel.info)
                    if not lima_client._saved:
                        return

                    if not writer.is_internal:
                        use_references, _ = self._scan_writer._save_reference_mode(
                            lima_client.file_format
                        )
                    else:
                        use_references = False

                    if use_references:
                        lima_client = lima_client_factory(
                            self._data_store, channel.info, ref_mode=True
                        )

                    self._lima_clients[channel.name] = (lima_client, use_references)

                last_status = data[-1]
                lima_client.update(**last_status)  # Only the last status is relevant
                self._add_lima_data(writer, lima_client, use_references)
            elif channel.data_type == types.ChannelDataType.NUMERIC_DATA:
                writer.add_internal(data)
            else:
                raise NotImplementedError(
                    f"Channel type {channel.data_type} not supported"
                )
        finally:
            if not npoints_before and writer.npoints:
                self._add_dataset_links(subscan, channel, writer)

    def _add_lima_data(
        self, writer: Hdf5DatasetWriter, lima_client, use_references: bool
    ) -> None:
        from_index = writer.npoints
        if use_references:
            _, file_format = self._scan_writer._save_reference_mode(
                lima_client.file_format
            )
            refs = lima_client[from_index:]
            if file_format == "hdf5":
                urls = [
                    (ref.file_path + "::" + ref.data_path, ref.index) for ref in refs
                ]
            else:
                urls = [(ref.file_path, ref.index) for ref in refs]
            writer.add_external(urls, file_format)
        else:
            # Read images from lima or file and add them to the writer
            images = lima_client[from_index:]
            if len(images) > 0:
                writer.add_internal(images)

    def _flush_pending_data(self, subscan: SubScanInfo):
        subscan.logger.info("Flush pending data")
        pending_data = self._subscan_to_pending_data[subscan.identifier]
        for pending in pending_data.values():
            self._force_add_channel_data(pending.channel, pending.get_data())
        pending_data.clear()

    def _add_dataset_links(
        self,
        subscan: SubScanInfo,
        channel: types.Channel,
        writer: Union[Hdf5DatasetWriter, Hdf5ScanReferenceWriter],
    ):
        if channel.data_type == types.ChannelDataType.SCAN_REFERENCE:
            return
        self._scan_writer._add_to_measurement_group(subscan, writer)
        if writer.device_type in ("positioner", "positionergroup"):
            self._scan_writer._add_to_positioners_group(subscan, writer)


class _PendingNumericData:
    def __init__(self, channel: types.Channel, data: typing.NumericDataType):
        self._buffer = list(data)
        self._channel = channel
        self._as_numpy = isinstance(data, numpy.ndarray)

    @property
    def channel(self) -> types.Channel:
        return self._channel

    def add_data(self, data: typing.NumericDataType) -> None:
        self._buffer.extend(data)

    def get_data(self) -> typing.NumericDataType:
        if self._as_numpy:
            detector_shapes = {arr.shape[1:] for arr in self._buffer}
            if len(detector_shapes) == 1:
                return numpy.array(self._buffer)
            # This is a ragged array (e.g. diode samples)
        return self._buffer


class _PendingLimaData:
    def __init__(self, channel: types.Channel, data: typing.LimaStatusType):
        self._data = data
        self._channel = channel

    @property
    def channel(self) -> types.Channel:
        return self._channel

    def add_data(self, data: typing.LimaStatusType) -> None:
        self._data = data

    def get_data(self) -> typing.LimaStatusType:
        return self._data


class _PendingScanReferenceData:
    def __init__(self, channel: types.Channel, data: typing.ScanReferenceType):
        self._data = data
        self._channel = channel

    @property
    def channel(self) -> types.Channel:
        return self._channel

    def add_data(self, data: typing.ScanReferenceType) -> None:
        self._data.extend(data)

    def get_data(self) -> typing.ScanReferenceType:
        return self._data


_CHANNEL_TYPE_TO_PENDING_CLASSES = {
    types.ChannelDataType.NUMERIC_DATA: _PendingNumericData,
    types.ChannelDataType.LIMA_STATUS: _PendingLimaData,
    types.ChannelDataType.SCAN_REFERENCE: _PendingScanReferenceData,
}



import threading
from contextlib import contextmanager
from typing import Hashable, List, Generator

try:
    from gevent.monkey import is_anything_patched
    import gevent.lock
except ImportError:

    def is_anything_patched():
        return False


class SharedLockPool:
    """
    Allows to acquire locks identified by key (hashable type) recursively.
    """

    def __init__(self, timeout: float = -1):
        self.__locks = dict()
        if is_anything_patched():
            self.__locks_mutex = gevent.lock.Semaphore(value=1)
            self.__lock_class = gevent.lock.RLock
        else:
            self.__locks_mutex = threading.Lock()
            self.__lock_class = threading.RLock
        self._timeout = timeout

    def __len__(self) -> int:
        with self.__locks_mutex:
            return len(self.__locks)

    def keys(self) -> List[Hashable]:
        with self.__locks_mutex:
            return list(self.__locks)

    @contextmanager
    def acquire(self, key: Hashable) -> Generator[None, None, None]:
        lock = self.__get_lock(key)
        try:
            if not lock.acquire(timeout=self._timeout):
                raise TimeoutError(f"could not lock {key}")
            try:
                yield
            finally:
                lock.release()
        finally:
            self.__pop_lock(key)

    def __get_lock(self, key: Hashable) -> threading.RLock:
        with self.__locks_mutex:
            lock, count = self.__locks.get(key, (None, 0))
            if lock is None:
                lock = self.__lock_class()
            self.__locks[key] = lock, count + 1
            return lock

    def __pop_lock(self, key: Hashable) -> None:
        with self.__locks_mutex:
            lock, count = self.__locks.pop(key, (None, 0))
            count -= 1
            if count > 0:
                self.__locks[key] = lock, count



import enum
import time
import logging
import datetime
from contextlib import contextmanager
from typing import Optional, Dict, Union, Tuple, Generator

import numpy

from blissdata.redis_engine.scan import Scan
from blissdata.redis_engine.scan import ScanState
from blissdata.redis_engine.stream import Stream
from blissdata.redis_engine.stream import StreamingClient
from blissdata.redis_engine.exceptions import EndOfStream

from ..parameters import default_saveoptions
from ..writer import types
from ..utils.periodic import PeriodicTask
from ..writer.datasets import format_bytes
from ..writer.main import ScanWriterWithState
from ..writer.capture import capture_exceptions
from ..writer.typing import ChannelDataType
from .base_subscriber import BaseSubscriber


_logger = logging.getLogger(__name__)


class NexusScanSubscriber(BaseSubscriber):
    @enum.unique
    class _STATE(enum.IntEnum):
        """
        * INIT: initializing (not listening to events yet)
        * ON: listening to events
        * OFF: not listening to events and resources released
        * FAULT: not listening to events due to exception
        """

        INIT = enum.auto()
        ON = enum.auto()
        FAULT = enum.auto()
        OFF = enum.auto()

    _ALLOWED_STATE_TRANSITIONS = {
        _STATE.INIT: [_STATE.INIT, _STATE.ON, _STATE.OFF, _STATE.FAULT],
        _STATE.ON: [_STATE.OFF, _STATE.FAULT],
        _STATE.OFF: [_STATE.OFF, _STATE.FAULT],
        _STATE.FAULT: [],
    }

    def __init__(
        self,
        scan: Scan,
        parent_logger=None,
        start_semaphore_file: Optional[str] = None,
        **kw,
    ) -> None:
        if parent_logger is None:
            parent_logger = _logger
        kw = {**default_saveoptions(), **kw}
        resource_profiling = kw.pop("resource_profiling")
        super().__init__(
            name=scan.info["name"],
            parent_logger=parent_logger,
            resource_profiling=resource_profiling,
            start_semaphore_file=start_semaphore_file,
        )

        self._scan = scan
        self._writer = ScanWriterWithState(
            scan.key, self._name, parent_logger=self._logger, **kw
        )
        self._scan_writer = self._writer.scan_writer
        self._channels: Dict[str, types.Channel] = dict()
        self._yield_period: float = 1
        self._minimum_read_period: float = 0.1
        self._periodic_tasks = [
            PeriodicTask(self._log_progress, 5),
        ]

    @property
    def scan_writer(self):
        return self._scan_writer

    def _run(self) -> None:
        super()._run()
        with self._scan_saving_context() as saving:
            if not saving:
                return
            for output in self._iter_data():
                for stream, (first_index, data) in output.items():
                    channel = self._compile_channel(stream)
                    channel_fully_defined = self._writer.add_channel_data(channel, data)
                    if not channel_fully_defined:
                        self._channels.pop(stream.name)
                self._execute_periodic_tasks()

    @contextmanager
    def _scan_saving_context(self) -> Generator[bool, None, None]:
        """Returns `True` when saving is required"""
        self._wait_scan_state(ScanState.PREPARED)
        if not self._writer.initialize(self._scan.info, self._scan._data_store):
            self._logger.info("saving is disabled")
            yield False
            return

        failed = False
        try:
            yield True
        except BaseException:
            failed = True
            raise
        finally:
            with capture_exceptions() as capture:
                with capture():
                    self._wait_scan_state(ScanState.CLOSED)
                with capture():
                    self._writer.finalize(self._scan.info, failed)
                with capture():
                    self._log_progress(
                        f"Finished writing to '{self._scan_writer.scan_info.filename}'"
                    )

    def _iter_data(
        self,
    ) -> Generator[Dict[Stream, Tuple[int, ChannelDataType]], None, None]:
        """Yields stream data until end of scan or `stop` method called."""
        client = StreamingClient(self._scan.streams)
        while not self._stop_requested:
            try:
                output = client.read(timeout=self._yield_period)
            except EndOfStream:
                break

            t0 = time.time()
            yield output
            # Make sure we don't call `client.read` too fast as
            # this stresses the Redis server too much.
            dt = max(time.time() - t0 - self._minimum_read_period, 0)
            time.sleep(dt)

        if self._stop_requested:
            raise RuntimeError("Stop requested")

    def _wait_scan_state(self, state: ScanState) -> None:
        """Wait until the scan state is reached or a stop is requested, in which case `RuntimeError` is raised."""
        while not self._stop_requested and self._scan.state < state:
            self._scan.update(timeout=self._yield_period)
        if self._stop_requested:
            raise RuntimeError("Stop requested")

    def _execute_periodic_tasks(self):
        for task in self._periodic_tasks:
            task.execute()

    def _compile_channel(self, stream: Stream) -> types.Channel:
        channel = self._channels.get(stream.name)
        if channel is not None:
            return channel
        if stream.encoding["type"] == "numeric":
            channel_type = types.ChannelDataType.NUMERIC_DATA
            channel_info = {
                "shape": stream.encoding["shape"],
                "dtype": numpy.dtype(stream.encoding["dtype_str"]),
                **stream.info,
            }
        elif stream.encoding["type"] == "json" and "lima" in stream.info["format"]:
            channel_type = types.ChannelDataType.LIMA_STATUS
            channel_info = dict(stream.info)
        elif stream.encoding["type"] == "json" and "subscan" in stream.info["format"]:
            channel_type = types.ChannelDataType.SCAN_REFERENCE
            channel_info = dict(stream.info)
        else:
            raise NotImplementedError(
                f'Stream not supported: encoding={stream.encoding["type"]}, format={stream.info["format"]}'
            )
        shape = channel_info.get("shape", None)
        if shape is not None:
            channel_info["shape"] = tuple(shape)
        channel = types.Channel(
            name=stream.name, data_type=channel_type, info=channel_info
        )
        self._channels[stream.name] = channel
        return channel

    def done(self, seconds: float = 0) -> bool:
        """Returns `True` when the subscriber finished x seconds ago."""
        if self.is_alive():
            return False
        if self._end_time is None:
            return False
        timediff = datetime.datetime.now().astimezone() - self._end_time
        return timediff >= datetime.timedelta(seconds=seconds)

    def get_progress(self) -> Dict[str, str]:
        """Progress of each subscan"""
        pdict = dict()

        def getprogress(tpl):
            return tpl[1]

        for _, subscan in self._scan_writer.scan_info.iter_subscans():
            lst = list()
            for _, dproxy in list(subscan.dataset_items()):
                if dproxy is not None:
                    lst.append(dproxy.progress_string)
            if lst:
                pstr = f"{min(lst, key=getprogress)[0]}-{max(lst, key=getprogress)[0]}"
            else:
                pstr = "0pts-0pts"
            pdict[subscan.name] = pstr
        return pdict

    def get_progress_info(self) -> Dict[str, str]:
        """Progress info of each subscan"""
        data = format_bytes(self._get_current_bytes())
        state = self.state.name
        pdict = self.get_progress()

        start = self.start_time
        end = self.end_time
        init = self.init_time

        # Delay between scan start and writer instantiation
        startdelay1 = _timediff(init, self._scan_writer.scan_info.scan_start_time)

        # Delay between writer instantiation and start of writing
        startdelay2 = _timediff(start, init)

        # Delay between end of scan and end of writing (or now)
        if end is not None:
            enddelay = _timediff(end, self._scan_writer.scan_info.scan_end_time)
        else:
            enddelay = _timediff(self._now, self._scan_writer.scan_info.scan_end_time)

        # Time between start of writing and end of writing (or now)
        duration = self.get_duration()

        # Format output
        if start is not None:
            start = start.strftime("%Y-%m-%d %H:%M:%S")
        else:
            start = "not started"
        if end is not None:
            end = end.strftime("%Y-%m-%d %H:%M:%S")
        else:
            end = "not finished"
        template = f"{state},{{}},{data},duration: {duration},start: {start} (delay: {startdelay1} + {startdelay2}),end: {end} (delay: {enddelay})"
        if pdict:
            return {name: template.format(s) for name, s in pdict.items()}
        return {"": template.format("0pts-0pts")}

    def _get_overall_progress(self) -> str:
        """Progress of the scan"""
        pdict = self.get_progress()
        if not pdict:
            return "0pts-0pts"
        if len(pdict) == 1:
            return next(iter(pdict.values()))
        plist = [f"{name}:{s}" for name, s in pdict.items()]
        return " ".join(plist)

    def _log_progress(self, msg=None):
        data = format_bytes(self._get_current_bytes())
        progress = self._get_overall_progress()
        duration = self.get_duration()
        if msg:
            self._logger.info("%s (%s %s %s)", msg, progress, data, duration)
        else:
            self._logger.info(" %s %s %s", progress, data, duration)

    def _get_current_bytes(self) -> int:
        """Total bytes (data-only) of all subscans"""
        nbytes = 0
        for _, subscan in self._scan_writer.scan_info.iter_subscans():
            for _, dproxy in list(subscan.dataset_items()):
                if dproxy is not None:
                    nbytes += dproxy.current_bytes
        return nbytes


def _timediff(
    tend: Union[str, datetime.datetime, None],
    tstart: Union[str, datetime.datetime, None],
) -> str:
    if isinstance(tend, str):
        tend = datetime.datetime.fromisoformat(tend)
    if isinstance(tstart, str):
        tstart = datetime.datetime.fromisoformat(tstart)
    if tend is None or tstart is None:
        return "NaN"
    if tend >= tstart:
        return str(tend - tstart)
    return f"-{tstart-tend}"



"""Mapping between blissdata channels and devices info"""

import re
from typing import Dict

from . import devices


def normalize_nexus_name(name: str):
    # TODO: could cause unique names to become non-unique ...
    return re.sub("[^a-zA-Z0-9_]+", "_", name)


def get_device_info(channel_name: str, subscan_devices: Dict) -> Dict:
    """Device information from which the NeXus structure
    of the channel data and metadata is derived."""
    device = subscan_devices.get(channel_name, None)
    if device is None:
        device = devices.update_device(subscan_devices, channel_name)
    return device


def get_primary_dataset_path(channel_name: str, subscan_devices: Dict) -> str:
    """Primary HDF5 dataset in which the channel data is saved."""
    device = get_device_info(channel_name, subscan_devices)
    if device["device_name"]:
        group = normalize_nexus_name(device["device_name"])
        dataset = normalize_nexus_name(device["data_name"])
        return f"instrument/{group}/{dataset}"
    dataset = normalize_nexus_name(device["data_name"])
    return f"measurement/{dataset}"



import logging
from collections import abc
import datetime
from typing import (
    Dict,
    Optional,
    Any,
    Union,
    MutableMapping,
    Iterator,
    Tuple,
    List,
    Set,
    Generator,
)

import h5py

from ..nexus import devices as nexus_devices
from ..nexus import mapping as nexus_mapping
from ..io import nexus
from .datasets import Hdf5DatasetWriter
from .references import Hdf5ScanReferenceWriter
from . import scan_url_info
from ..utils.logging_utils import CustomLogger
from ..utils.array_order import Order


_logger = logging.getLogger(__name__)


class _NotProvided:
    pass


_NOT_PROVIDED = _NotProvided()


class _BaseInfo(abc.MutableMapping):
    def __init__(
        self,
        identifier: str,
        name: str,
        info: Optional[MutableMapping] = None,
        parent_info: Optional["_BaseInfo"] = None,
        parent_logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None,
    ) -> None:
        self._name = name
        self._identifier = identifier
        if info is None:
            info = dict()
        self._info = info
        self._parent_info = parent_info

        if parent_logger is None:
            parent_logger = _logger
        self._logger = parent_logger

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def name(self) -> str:
        return self._name

    @property
    def logger(self) -> CustomLogger:
        return self._logger

    def __repr__(self) -> str:
        return self._identifier

    def __str__(self) -> str:
        return self._name

    def __getitem__(self, key: str) -> Any:
        return self._info[key]

    def __setitem__(self, key: str, value: Any):
        self._info[key] = value

    def __delitem__(self, key: str) -> None:
        del self._info[key]

    def __iter__(self) -> Iterator:
        return iter(self._info)

    def __len__(self) -> int:
        return len(self._info)

    def get(self, key: str, default: Any = _NOT_PROVIDED) -> Any:
        """When a key is missing it first tries from the parent. If that
        fails well it returns the default. Default value (when explicitly provided)
        and parent value are cached in this object and in the parent.
        """
        try:
            value = self[key]
        except KeyError:
            if self._parent_info is not None:
                value = self._parent_info.get(key, default=default)
                self[key] = value
            elif default is _NOT_PROVIDED:
                value = None
            else:
                value = default
                self[key] = value
        return value


class SubScanInfo(_BaseInfo):
    def __init__(
        self,
        identifier: str,
        name: str,
        scan_info: "ScanInfo",
        info: Optional[MutableMapping] = None,
        parent_logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None,
    ) -> None:
        if parent_logger is None:
            parent_logger = _logger
        parent_logger = CustomLogger(parent_logger, name)
        super().__init__(
            identifier,
            name,
            info=info,
            parent_info=scan_info,
            parent_logger=parent_logger,
        )
        self._dataset_writers: Dict[str, Hdf5DatasetWriter] = dict()
        self._scanref_writers: Dict[str, Hdf5ScanReferenceWriter] = dict()

    def get_reference_writer(
        self, identifier: str
    ) -> Optional[Hdf5ScanReferenceWriter]:
        return self._scanref_writers.get(identifier)

    def get_dataset_writer(self, identifier: str) -> Optional[Hdf5DatasetWriter]:
        return self._dataset_writers.get(identifier)

    def reference_items(self) -> Iterator[Tuple[str, Hdf5ScanReferenceWriter]]:
        return self._scanref_writers.items()

    def dataset_items(self) -> Iterator[Tuple[str, Hdf5DatasetWriter]]:
        return self._dataset_writers.items()

    def add_reference(
        self, identifier: str, filename: str, nxroot
    ) -> Hdf5ScanReferenceWriter:
        ref = self._scanref_writers.get(identifier)
        if ref is not None:
            raise RuntimeError(f"Reference '{identifier}' already added")
        ref = Hdf5ScanReferenceWriter(
            filename=filename,
            parent="/",
            filecontext=nxroot,
            nreferences=0,
            parent_logger=self.logger,
        )
        self._scanref_writers[identifier] = ref
        self.logger.debug("Created reference writer %s", ref)
        return ref

    def add_dataset(
        self,
        identifier: str,
        filename: str,
        nxroot: Generator[Optional[nexus.nxRoot], None, None],
        parent: h5py.Group,
        device: dict,
        scan_shape: Tuple[int],
        scan_save_shape: Tuple[int],
        saveorder: Order,
        chunk_options: dict,
        data_info: dict,
    ) -> Hdf5DatasetWriter:

        dset = self._dataset_writers.get(identifier)
        if dset is not None:
            raise RuntimeError(f"Dataset '{identifier}' already added")
        dset = Hdf5DatasetWriter(
            filename=filename,
            parent=parent,
            filecontext=nxroot,
            device=device,
            scan_shape=scan_shape,
            scan_save_shape=scan_save_shape,
            saveorder=saveorder,
            publishorder=saveorder,
            parent_logger=self.logger,
            chunk_options=chunk_options,
            **data_info,
        )
        self._dataset_writers[identifier] = dset
        self.logger.debug("Created dataset writer %s", identifier)
        return dset

    def flush(self) -> None:
        for writer in self._dataset_writers.values():
            if writer.npoints:
                writer.flush()
        for writer in self._scanref_writers.values():
            if writer.npoints:
                writer.flush()


class ScanInfo(_BaseInfo):
    def __init__(
        self,
        identifier: str,
        name: str,
        info: Optional[MutableMapping] = None,
        configurable: bool = True,
        flat: bool = True,
        short_names: bool = True,
        multivalue_positioners: bool = False,
        writer_name: Optional[str] = None,
        parent_logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None,
    ) -> None:
        super().__init__(identifier, name, info=info, parent_logger=parent_logger)

        # Save options
        self._configurable = configurable
        self._short_names = short_names
        self._multivalue_positioners = multivalue_positioners
        self._flat = flat
        self._save_order = Order("C")
        self._writer_name = writer_name

        # Runtime state
        self._filename: Optional[str] = None
        self._devices: Dict[
            str, Dict[str, str]
        ] = dict()  # keys: subscan names (from acquisition chain)
        self._subscans: Dict[
            str, SubScanInfo
        ] = dict()  # keys: subscan identifiers (added)

    def get_save_options(self) -> dict:
        return {
            "configurable": self._configurable,
            "short_names": self._short_names,
            "multivalue_positioners": self._multivalue_positioners,
            "flat": self._flat,
        }

    def get(self, key, default=None, subscan: Optional[SubScanInfo] = None):
        if subscan is None:
            return super().get(key, default=default)
        if subscan is None:
            return super().get(key, default=default)
        return subscan.get(key, default=default)

    def get_subscan(self, identifier: str) -> "SubScanInfo":
        return self._subscans[identifier]

    def add_subscan(
        self, identifier: str, name: str, info: Optional[MutableMapping] = None
    ) -> "SubScanInfo":
        subscan = self._subscans.get(identifier)
        if subscan is not None:
            raise RuntimeError(f"Subscan '{identifier}' already added")
        subscan = SubScanInfo(
            identifier, name, self, info=info, parent_logger=self._logger
        )
        self._subscans[identifier] = subscan
        return subscan

    def iter_subscans(self) -> Iterator[Tuple[str, SubScanInfo]]:
        yield from self._subscans.items()

    @property
    def _expected_subscans(self):
        """
        Subscan names for which there are devices defined.
        Ordered according to position in acquisition chain.
        """
        return list(sorted(self.devices.keys()))

    def nxentry_name(self, subscan: SubScanInfo) -> Optional[str]:
        """Name of the NXentry associated with a subscan"""
        if self.scan_number is None:
            self._log_h5creation("scan number not specified")
            return None
        try:
            chain_index = self._expected_subscans.index(subscan.name)
        except ValueError:
            self._log_h5creation(
                f"subscan '{subscan.name}' not in the acquisition chain"
            )
            return None
        return f"{self.scan_number}.{chain_index + 1}"

    def nxentry_create_args(self):
        """Arguments for NXentry creation"""
        start_time = self.get("start_time")
        if not start_time:
            self._log_h5creation("'start_time' not specified")
            return None
        title = self.get("title")
        if not title:
            self._log_h5creation("'title' not specified")
            return None
        start_time = datetime.datetime.fromisoformat(start_time)
        nxdict = {"title": title}
        kwargs = {"start_time": start_time, "nxdict": nxdict}
        return kwargs

    def _log_h5creation(self, reason: str) -> None:
        self._logger.debug("HDF5 group not created ({%s})", reason)

    @property
    def devices(self):
        """Maps subscan name to a dictionary of devices,
        which maps fullname to device info. Ordered
        according to position in acquisition chain.
        """
        if not self._devices:
            config_devices = self.config_writer.get("devices", dict())
            self._devices = nexus_devices.device_info(
                config_devices,
                self,
                scan_ndim=self.scan_ndim(),
                short_names=self._short_names,
                multivalue_positioners=self._multivalue_positioners,
            )
        return self._devices

    def get_device_info(
        self, subscan: SubScanInfo, channel_name: Optional[str]
    ) -> Optional[Dict]:
        if not channel_name:
            return None
        subdevices = self.devices[subscan.name]
        return nexus_mapping.get_device_info(channel_name, subdevices)

    @property
    def filename(self) -> Optional[str]:
        if self._filename is None:
            self._filename = scan_url_info.scan_filename(self, raw=True)
        return self._filename

    def get_urls(self) -> List[str]:
        filename = self.filename
        if not filename:
            return list()
        ret = dict()
        for _, subscan in self.iter_subscans():
            name = self.nxentry_name(subscan)
            ret[name.split(".")[-1]] = filename + "::/" + name
        return [v for _, v in sorted(ret.items())]

    def get_subscan_url(self, subscan: SubScanInfo) -> Optional[str]:
        filename = self.filename
        if not filename:
            return None
        return f"{self.filename}::/{self.nxentry_name(subscan)}"

    def get_master_filenames(self) -> Dict[str, str]:
        info = self.config_writer
        if info:
            return info.get("masterfiles", dict())
        filename = self.filename
        master_filename = self.get("filename")
        if filename == master_filename:
            return dict()
        return {"dataset": master_filename}

    def saving_enabled(self) -> bool:
        """Saving intended for this scan?"""
        if not self.get("save", False):
            return False
        if not self._writer_name:
            return True
        return self.get("data_writer", None) == self._writer_name

    @property
    def scan_number(self) -> Optional[int]:
        return self.get("scan_nb")

    @property
    def scan_start_time(self) -> Optional[str]:
        return self.get("start_time")

    @property
    def scan_end_time(self) -> Optional[str]:
        return self.get("end_time")

    @property
    def writer_options(self) -> dict:
        return self.get("writer_options", dict())

    @property
    def chunk_options(self) -> dict:
        return self.writer_options.get("chunk_options", dict())

    @property
    def config_writer(self) -> dict:
        if self._configurable:
            return self.get("nexuswriter", dict())
        return dict()

    @property
    def instrument_info(self) -> dict:
        return self.config_writer.get("instrument_info", dict())

    @property
    def config_technique(self) -> dict:
        return self.config_writer.get("technique", dict())

    @property
    def technique_name(self) -> str:
        return self.config_technique.get("name", "")

    @property
    def applications(self) -> dict:
        return self.config_technique.get("applications", dict())

    @property
    def positioner_info(self) -> dict:
        return self.get("positioners", dict())

    @property
    def motors(self) -> List[str]:
        return list(self.positioner_info.get("positioners_start", dict()).keys())

    def scan_ndim(self, subscan: Optional[str] = None) -> int:
        """Number of dimensions of the scan or subscan (0 means a scalar scan but currently
        no scan has this)
        """
        default = 1  # 1D scan by default
        # TODO: currently "data_dim" is not published in the subscan but in the scan.
        ndim = self.get("data_dim", default=default, subscan=subscan)
        if ndim <= 1:
            return ndim
        if all(
            self.get(f"npoints{i}", default=0, subscan=subscan)
            for i in range(1, ndim + 1)
        ):
            return ndim
        return 1

    def scan_size(self, subscan: Optional[str] = None) -> int:
        """Number of points in the subscan (0 means variable-length)"""
        # TODO: currently "npoints" is not published in the subscan but in the scan.
        return self.get("npoints", default=0, subscan=subscan)

    def scan_shape(self, subscan: Optional[str] = None) -> Tuple[int]:
        """Shape of the subscan (empty means a scalar scan like a ct).
        Dimensions with size 0 are variable (can only occur for 1D scans).
        """
        ndim = self.scan_ndim(subscan)
        if ndim == 0:
            # A scalar scan like a ct
            return tuple()
        if ndim == 1:
            # A 1D scan like a loopscan (fixed) or timescan (variable)
            return (self.scan_size(subscan),)
        # An nD scan like a mesh
        # Fast axis first
        s = tuple(
            self.get(f"npoints{i}", default=0, subscan=subscan)
            for i in range(1, ndim + 1)
        )
        if self._save_order.order == "C":
            # Fast axis last
            s = s[::-1]
        return s

    def scan_save_shape(self, subscan: Optional[str] = None) -> Tuple[int]:
        """Potentially flattened `scan_shape`"""
        if self._flat:
            if self.scan_ndim(subscan) == 0:
                return tuple()
            else:
                return (self.scan_size(subscan),)
        else:
            return self.scan_shape(subscan)

    def scan_save_ndim(self, subscan: Optional[str] = None) -> int:
        """Potentially flattened scan_ndim"""
        return len(self.scan_save_shape(subscan))

    @property
    def detector_ndims(self) -> Set[int]:
        """All detector dimensions"""
        ret = set()
        for _, subscan in self.iter_subscans():
            for _, dproxy in list(subscan.dataset_items()):
                if dproxy is not None:
                    ret.add(dproxy.detector_ndim)
        return ret

    @property
    def plots(self) -> Dict[str, dict]:
        """NXdata signals"""
        adict = self._flintplots
        if self._configurable:
            adict.update(self.config_technique.get("plots", {}))
        else:
            for ndim in self.detector_ndims:
                adict["plot{}D".format(ndim)] = {"ndim": ndim}
        return adict

    @property
    def _flintplots(self) -> dict:
        display_extra = self.get("display_extra", default={})
        items = display_extra.get("displayed_channels", None)
        if items is None:
            items = display_extra.get("plotselect", None)
        if items:
            return {"plotselect": {"items": items, "grid": True}}
        return dict()

    @property
    def plotselect(self) -> str:
        if not self._configurable or self._flintplots:
            return ""
        return self.config_technique.get("plotselect", "")



import pytest
import os
import numpy
import h5py
import fabio
import itertools

try:
    import gzip
except ImportError:
    gzip = None
from fabio.edfimage import EdfImage
from blisswriter.io import h5_external
from blisswriter.utils.array_order import Order


def test_h5ext_edf(tmpdir):
    scan_shape = 2, 3
    image_shape = 5, 4
    files = edf_files(tmpdir, scan_shape, image_shape)
    for data, filename, indices, order, compression in files:
        if indices:
            filenames = [(filename, i) for i in indices]
        else:
            filenames = filename
        # Save as HDF5:
        if compression != "NONE":
            with pytest.raises(RuntimeError):
                kwargs = h5_external.add_edf_arguments(filenames)
            continue
        else:
            kwargs = h5_external.add_edf_arguments(filenames)
            h5_external.finalize(kwargs, shape=scan_shape, addorder=order)
        # TODO: external datasets do not support relative paths
        # kwargs['external'] = [(os.path.relpath(tpl[0], str(tmpdir)),) + tpl[1:]
        #                      for tpl in kwargs['external']]
        filename = os.path.join(str(tmpdir), "out.h5")
        with h5py.File(filename, mode="w") as f:
            data2 = f.create_dataset("data", **kwargs)
            # Check HDF5
            numpy.testing.assert_array_equal(data, data2[()])


def test_h5ext_edf_append(tmpdir):
    n = 3
    scan_shape_org = 2, 3
    image_shape = 5, 4
    scan_shape = 2, 3 * n
    enpoints = numpy.prod(scan_shape)
    files = edf_files(tmpdir, scan_shape_org, image_shape)
    for data, filename, indices, order, compression in files:
        if indices:
            filenames = [(filename, indices)]
        else:
            filenames = [filename]
        if compression != "NONE":
            continue
        for shape in tuple(), scan_shape:
            kwargs1 = {}
            for i in range(n):
                kwargs1 = h5_external.add_edf_arguments(filenames, createkwargs=kwargs1)
            h5_external.finalize(kwargs1, shape=shape, addorder=order)
            kwargs2 = h5_external.add_edf_arguments(filenames * n)
            h5_external.finalize(kwargs2, shape=shape, addorder=order)
            assert kwargs1 == kwargs2
            npoints = numpy.prod(kwargs1["shape"]) // numpy.prod(image_shape)
            assert npoints == enpoints


def edf_files(tmpdir, scan_shape, image_shape):
    nimages = numpy.prod(scan_shape)
    shape = scan_shape + image_shape
    fshape = (numpy.prod(scan_shape),) + image_shape
    fdata = numpy.arange(numpy.prod(shape))
    # Data save in C order
    data = fdata.reshape(shape, order="C")
    # Data publication in any order
    order = Order("C"), Order("F"), Order("C", caxes="all"), Order("F", caxes="all")
    compression = "NONE", "GZIP"
    withindices = False, True
    options = itertools.product(order, compression, withindices)
    for i, (order, compression, withindices) in enumerate(options):
        # Save as EDF:
        filename = os.path.join(str(tmpdir), "out{}.edf".format(i))
        data1 = order.reshape(data, fshape)
        edf = None
        for img in data1:
            if compression == "GZIP":
                if gzip is None:
                    continue
            # header = {'COMPRESSION': compression}
            header = None
            if edf is None:
                edf = EdfImage(data=img, header=header)
            else:
                edf.append_frame(data=img, header=header)
        if compression == "GZIP":
            filename += ".gz"
            f = gzip.GzipFile(filename, "wb")
        else:
            f = filename
        edf.write(f)
        # Check EDF:
        edf = fabio.open(filename)
        data2 = numpy.array([frame.data for frame in edf.frames()])
        numpy.testing.assert_array_equal(data1, data2)
        if withindices:
            indices = list(range(nimages))
        else:
            indices = None
        yield data, filename, indices, order, compression



import sys
from blisswriter.tango.servers import NexusWriter
from blisswriter.utils.logging_utils import add_cli_args
from blisswriter.utils.log_levels import tango_cli_slog_level
from blisswriter.utils.log_levels import add_tango_cli_args
from .utils import config_root_logger


def run(server, instance, log_level):
    """
    :param str server: device server name
    :param str instance: device server instance name
    :param str log_level:
    :returns Util:
    """
    verbose = tango_cli_slog_level.get(log_level, 0)
    if verbose:
        verbose = "-v{:d}".format(verbose)
        serverargs = [server, instance, verbose]
    else:
        serverargs = [server, instance]
    return NexusWriter.main(args=serverargs)


def main(argv=None):
    import argparse

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Start a Tango server for saving Bliss data in HDF5"
    )
    parser.add_argument(
        "instance",
        type=str,
        default="nexuswriters",
        help="Server instance name ('nexuswriters' by default)",
    )
    parser.add_argument(
        "--server",
        type=str,
        default="nexuswriter",
        help="Server name ('nexuswriter' by default)",
    )
    add_cli_args(parser)
    add_tango_cli_args(parser)
    args = parser.parse_args(argv[1:])
    config_root_logger(args)
    run(args.server, args.instance, args.log_tango)


if __name__ == "__main__":
    sys.exit(main())



import pytest
import numpy
import itertools

from blisswriter.utils import data_merging
from blisswriter.utils.array_order import Order

from .helpers.nxw_test_math import asproduct


@pytest.mark.parametrize("axis", (0, 1, -1))
@pytest.mark.parametrize("newaxis", (True, False))
@pytest.mark.parametrize("reshape", (False, True))
@pytest.mark.parametrize("order", (Order("C"), Order("F")))
@pytest.mark.parametrize("advanced", (False, True))
def test_data_merging_general(axis, newaxis, reshape, order, advanced):
    shape = 3, 7
    dtype = float
    data = numpy.arange(numpy.prod(shape), dtype=dtype)
    data = data.reshape(shape)

    kwargs = {
        "axis": axis,
        "newaxis": newaxis,
        "order": order,
        "allow_advanced_indexing": advanced,
    }
    sources = [data] * 3
    shapes = [shape] * 3
    if not newaxis:
        extrashape = list(shape)
        extrashape[axis] *= 2
        extrashape = tuple(extrashape)
        extradata = numpy.arange(numpy.prod(extrashape), dtype=dtype)
        extradata = extradata.reshape(extrashape)
        sources.append(extradata)
        shapes.append(extrashape)
    if newaxis:
        edata = numpy.stack(sources, axis=axis)
    else:
        edata = numpy.concatenate(sources, axis=axis)
    if reshape:
        dshape = edata.shape
        it = asproduct(numpy.prod(dshape), 3, includeone=False)
        nshape = next(it)
        while nshape == dshape:
            nshape = next(it)
        edata = order.reshape(edata, nshape)
    else:
        nshape = None
    mshape, fillgen = data_merging.mergeGenerator(
        sources, shapes, shape=nshape, **kwargs
    )
    mdata = numpy.empty(shape=mshape, dtype=dtype)
    for ndarray, idx_generator in fillgen():
        for idxin, idxout in idx_generator():
            for idx in idxin, idxout:
                isadvanced = any(isinstance(x, numpy.ndarray) for x in idx)
                assert isadvanced == (advanced and reshape)
            mdata[idxout] = ndarray[idxin]
    numpy.testing.assert_array_equal(edata, mdata)


@pytest.mark.parametrize("order", (Order("C"), Order("F")))
@pytest.mark.parametrize("advanced", (False, True))
def test_data_merging_single(order, advanced):
    dshape = 13, 14
    scanshape1 = 6, 2, 3
    scanshape2 = (numpy.prod(scanshape1, dtype=int),)
    if order.order == "C":
        shape1 = scanshape1 + dshape
        shape2 = scanshape2 + dshape
    else:
        shape1 = dshape + scanshape1
        shape2 = dshape + scanshape2
    n = numpy.prod(shape1, dtype=int)
    data = numpy.arange(n, dtype=float).reshape(shape1)
    assert_single(data, shape2, order, advanced)
    data = numpy.arange(n, dtype=float).reshape(shape2)
    assert_single(data, shape1, order, advanced)


def assert_single(data, shape, order, advanced):
    edata = order.reshape(data, shape)
    mshape, fillgen = data_merging.mergeGenerator(
        [data],
        [data.shape],
        shape=shape,
        newaxis=False,
        order=order,
        allow_advanced_indexing=advanced,
    )
    assert shape == mshape
    mdata = numpy.empty(shape=mshape, dtype=data.dtype)
    for ndarray, idx_generator in fillgen():
        for idxin, idxout in idx_generator():
            mdata[idxout] = ndarray[idxin]
    numpy.testing.assert_array_equal(edata, mdata)


def test_data_merging_list_to_slice():
    idx = (numpy.array([4, 0, 1, 2, 3, 4, 0]), numpy.array([2, 3, 3, 3, 3, 3, 4]))
    assert_intlist_to_slice(idx)
    idx = (numpy.array([2, 1, 0, 1, 2, 3, 4, 0]), numpy.array([2, 2, 2, 3, 3, 3, 3, 4]))
    assert_intlist_to_slice(idx)
    idx = (
        numpy.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 6, 4, 2, 2, 2]),
        numpy.array([0, 1, 2, 3, 0, 1, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3]),
        numpy.array([0, 1, 0, 1, 0, 1, 2, 3, 4, 5, 6, 7, 1, 2, 3, 9, 8]),
    )
    assert_intlist_to_slice(idx)
    idx = (numpy.array([0]), numpy.array([0]), numpy.array([0]))
    assert_intlist_to_slice(idx)
    idx = (numpy.array([0, 0]), numpy.array([0, 2]), numpy.array([0, 4]))
    assert_intlist_to_slice(idx)
    idx = (numpy.array([0, 0, 1]), numpy.array([0, 2, 2]), numpy.array([0, 4, 4]))
    assert_intlist_to_slice(idx)
    idx = (numpy.array([1, 0, 0]), numpy.array([2, 0, 2]), numpy.array([4, 0, 4]))
    assert_intlist_to_slice(idx)
    combinations = []
    comb = [10], [20], [30]
    combinations.append(numpy.array(comb))
    comb = [1], [2], [3]
    combinations.append(numpy.array(comb))
    comb = [0] * 8, list(range(8)), list(range(1, 9))
    combinations.append(numpy.array(comb))
    comb = list(range(4)) + list(range(4)), [0] * 8, list(range(3)) + list(range(5))
    combinations.append(numpy.array(comb))
    for comb in itertools.permutations(combinations):
        idx = numpy.concatenate(comb, axis=1)
        assert_intlist_to_slice(idx)


def assert_intlist_to_slice(idx):
    for slicetwo in True, False:
        for allow_negative_stride in True, False:
            idx = numpy.asarray(idx)
            x = numpy.arange(idx.max() + 1)
            y = []
            for idx2 in data_merging.intListToSliceIndexing(
                idx, slicetwo=slicetwo, allow_negative_stride=allow_negative_stride
            ):
                if not allow_negative_stride:
                    assert all(
                        idx3.step > 0 for idx3 in idx2 if isinstance(idx3, slice)
                    )
                lst = [numpy.array(x[idx3]) for idx3 in idx2]
                n = max(arr.size for arr in lst)
                lst = [arr if arr.ndim else arr.repeat(n) for arr in lst]
                y.append(numpy.array(lst))
            idx2 = numpy.concatenate(y, axis=1)
            numpy.testing.assert_array_equal(idx, idx2)



from blisswriter.io import nexus
from blisswriter.writer.references import Hdf5ScanReferenceWriter


def test_dataset_proxy(tmpdir):
    mainfile = str(tmpdir.join("main.h5"))
    exfile = str(tmpdir.join("ex.h5"))
    with nexus.nxRoot(mainfile, mode="w") as nxroot:
        nexus.nxEntry(nxroot, "destination")
        nexus.nxEntry(nxroot, "references")
    with nexus.nxRoot(exfile, mode="w") as nxroot:
        nexus.nxEntry(nxroot, "references")
    refmain = [f"{mainfile}::/references/{i}" for i in range(0, 5)]
    refex = [f"{exfile}::/references/{i}" for i in range(5, 10)]
    references = refmain + refex
    nreferences = len(references)

    # Add references (no self-referencing)
    rproxy = Hdf5ScanReferenceWriter(
        filename=mainfile, parent="/destination", nreferences=nreferences
    )
    rproxy.add_references(references[: nreferences // 2])
    rproxy.add_references(references[nreferences // 2 :])
    assert rproxy.npoints == nreferences
    assert rproxy.complete
    with rproxy.open() as parent:
        references2 = [
            nexus.dereference(parent, name)
            for name in parent
            if nexus.isLink(parent, name)
        ]
    assert set(references) == set(references2)

    # Add references (self-referencing)
    rproxy = Hdf5ScanReferenceWriter(
        filename=mainfile, parent="/references", nreferences=nreferences
    )
    rproxy.add_references(references[: nreferences // 2])
    rproxy.add_references(references[nreferences // 2 :])
    assert rproxy.npoints == nreferences
    assert rproxy.complete
    with rproxy.open() as parent:
        references2 = [
            nexus.dereference(parent, name)
            for name in parent
            if nexus.isLink(parent, name)
        ]
    assert set(refex) == set(references2)



import os
import re
import logging
import datetime
from pprint import pformat
from contextlib import contextmanager
from importlib.metadata import version
from typing import (
    Dict,
    Optional,
    Union,
    Callable,
    Generator,
    Tuple,
    Iterator,
    Mapping,
    List,
    Any,
)

import h5py
from pyicat_plus.metadata.nexus import create_nxtreedict

from . import info
from . import datasets
from . import references
from . import plots
from ..utils.periodic import PeriodicTask
from ..io import nexus
from ..io import os_utils
from ..nexus.mapping import normalize_nexus_name


class ScanWriter:
    """Manage the saving of a Bliss scan in NeXus format"""

    def __init__(
        self,
        identifier: str,
        name: str,
        short_names: bool = True,
        expand_variable_length: bool = True,
        hold_file_open: bool = True,
        locking: bool = True,
        swmr: bool = False,
        flat: bool = True,
        multivalue_positioners: bool = False,
        allow_external_nonhdf5: bool = False,
        allow_external_hdf5: bool = True,
        copy_non_external: bool = False,
        required_disk_space: float = 200,
        recommended_disk_space: float = 1024,
        stack_mcas: bool = False,
        configurable: bool = True,
        writer_name: Optional[str] = None,
        parent_logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None,
        query_pool=None,
    ) -> None:
        # Save options
        self._expand_variable_length = expand_variable_length
        self._hold_file_open = hold_file_open
        self._locking = locking
        self._swmr = swmr
        self._allow_external_nonhdf5 = allow_external_nonhdf5
        self._allow_external_hdf5 = allow_external_hdf5
        self._copy_non_external = copy_non_external
        self._required_disk_space = required_disk_space
        self._recommended_disk_space = recommended_disk_space
        self._stack_mcas = stack_mcas
        self._applications: Dict[
            str, Callable[[info.SubScanInfo, str, Mapping], None]
        ] = {"appxrf": self._save_application_xrf}

        # Periodic tasks
        self._h5flush_task_period = 0.5  # seconds
        self._disk_check_period = 3  # seconds
        self._periodic_tasks = list()

        #
        self._query_pool = query_pool

        # Runtime state
        self._nxroot: Dict[str, nexus.nxRoot] = dict()  # scan and master files
        self._nxentry: Optional[h5py.Group] = None  # scan file only
        self._nxentry_created = False
        self._exception_is_fatal = False  # REMOVE
        if not hasattr(self, "_scan_info"):  # REMOVE when dropping the backport
            self._scan_info = info.ScanInfo(
                identifier,
                name,
                flat=flat,
                short_names=short_names,
                multivalue_positioners=multivalue_positioners,
                configurable=configurable,
                writer_name=writer_name,
                parent_logger=parent_logger,
            )

    def get_save_options(self) -> dict:
        return {
            **self._scan_info.get_save_options(),
            "expand_variable_length": self._expand_variable_length,
            "hold_file_open": self._hold_file_open,
            "locking": self._locking,
            "swmr": self._swmr,
            "allow_external_nonhdf5": self._allow_external_nonhdf5,
            "allow_external_hdf5": self._allow_external_hdf5,
            "copy_non_external": self._copy_non_external,
            "required_disk_space": self._required_disk_space,
            "recommended_disk_space": self._recommended_disk_space,
            "stack_mcas": self._stack_mcas,
        }

    @property
    def scan_info(self) -> info.ScanInfo:
        return self._scan_info

    def has_write_permissions(self):
        """This process has permission to write/create file and/or directory"""
        path = self._scan_info.filename
        if path:
            return os_utils.has_write_permissions(path)
        else:
            return True

    def disk_space_error(self) -> str:
        """Returns an error message when the disk space is running dangerously low."""
        return self._required_disk_space_msg(self._required_disk_space)

    def disk_space_warning(self) -> str:
        """Returns a warning message when the disk space is running low."""
        return self._required_disk_space_msg(self._recommended_disk_space)

    def _required_disk_space_msg(self, required_disk_space: int) -> str:
        path = self._scan_info.filename
        if not path:
            return ""
        if os_utils.has_required_disk_space(
            path, required_disk_space, query_pool=self._query_pool
        ):
            return ""
        return "Free disk space below {:.0f} MB".format(required_disk_space)

    def _check_required_disk_space(self) -> None:
        """
        :raises RuntimeError: when not enough space on disk
        """
        err_msg = self.disk_space_error()
        if err_msg:
            raise RuntimeError(err_msg)

    def _register_periodic_task(self, task: Callable[[], None], period: float = 0):
        self._periodic_tasks.append(PeriodicTask(task, period))

    def _register_periodic_write_tasks(self):
        self._register_periodic_task(
            self._check_required_disk_space, self._disk_check_period
        )
        for nxroot in self._nxroot.values():
            self._register_periodic_task(nxroot.flush, self._h5flush_task_period)

    def _process_periodic_tasks(self):
        """Execute tasks after succesfully processing a Redis event"""
        for task in self._periodic_tasks:
            task.execute()

    def add_dataset(
        self,
        subscan: info.SubScanInfo,
        identifier: str,
        parent: h5py.Group,
        device: dict,
        data_info: dict,
    ) -> datasets.Hdf5DatasetWriter:
        scan_shape = self._scan_info.scan_shape(subscan)
        scan_save_shape = self._scan_info.scan_save_shape(subscan)
        return subscan.add_dataset(
            identifier,
            self._scan_info.filename,
            self.nxroot,
            parent,
            device,
            scan_shape,
            scan_save_shape,
            self._scan_info._save_order,
            self._scan_info.chunk_options,
            data_info,
        )

    def add_reference(
        self, subscan: info.SubScanInfo, identifier: str
    ) -> references.Hdf5ScanReferenceWriter:
        return subscan.add_reference(identifier, self._scan_info.filename, self.nxroot)

    @contextmanager
    def nxroot(self, filename=None) -> Generator[Optional[nexus.nxRoot], None, None]:
        """
        Yields the NXroot instance (h5py.File) or None
        when information is missing
        """
        if not filename:
            filename = self._scan_info.filename
            if not filename:
                self._scan_info.logger.debug(
                    "HDF5 group not created yet ('filename' missing)"
                )
        nxroot = self._nxroot.get(filename, None)
        if nxroot is None:
            if filename:
                self._check_required_disk_space()
                try:
                    with nexus.nxRoot(filename, **self._nxroot_kwargs) as nxroot:
                        try:
                            self._nxroot[filename] = nxroot
                            yield nxroot
                        finally:
                            self._nxroot[filename] = None
                except OSError as e:
                    if nxroot is None and nexus.isLockedError(e):
                        self._exception_is_fatal = True
                        raise RuntimeError(nexus.lockedErrorMessage(filename)) from None
                    else:
                        raise
            else:
                yield None
        else:
            yield nxroot

    @property
    def _nxroot_kwargs(self) -> dict:
        if nexus.ISSUE_1641_FIXED:
            rootattrs = {
                "creator": "blissdata",
                "create_version": version("blissdata"),
                "publisher": self._scan_info.get("publisher", "bliss"),
                "publisher_version": self._scan_info.get("publisher_version", None),
            }
        else:
            rootattrs = {
                "creator": "blissdata",
                "create_version": version("blissdata"),
                "publisher_version": self._scan_info.get("publisher_version", None),
            }
        return {
            "mode": "a",
            "locking": self._locking,
            "swmr": self._swmr,
            "rootattrs": rootattrs,
        }

    @contextmanager
    def _modify_nxroot(
        self, filename=None
    ) -> Generator[Optional[nexus.nxRoot], None, None]:
        with self.nxroot(filename=filename) as nxroot:
            if nxroot is None:
                yield nxroot
            else:
                with nxroot.protect():
                    yield nxroot

    @contextmanager
    def nxentry(
        self, subscan: info.SubScanInfo
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxroot() as nxroot:
            if nxroot is None:
                yield None
                return
            if self._nxentry is None:
                nxentry = self._require_nxentry(nxroot, subscan)
                try:
                    self._nxentry = nxentry
                    yield nxentry
                finally:
                    self._nxentry = None
            else:
                yield self._nxentry

    def _require_nxentry(
        self, nxroot: nexus.nxRoot, subscan: info.SubScanInfo
    ) -> Optional[h5py.Group]:
        name = self._scan_info.nxentry_name(subscan)
        if not name:
            return None
        nxentry = nxroot.get(name, None)
        if nxentry is None:
            kwargs = self._scan_info.nxentry_create_args()
            if not kwargs:
                return None
            try:
                nxentry = nexus.nxEntry(nxroot, name, raise_on_exists=True, **kwargs)
            except nexus.NexusInstanceExists:
                self._exception_is_fatal = True
                raise RuntimeError(
                    f"More than one writer is writing to {self._scan_info.filename}"
                )
            else:
                self._nxentry_created = True
            url = repr(self._scan_info.get_subscan_url(subscan))
            name = repr(subscan.name)
            self._scan_info.logger.info("Start writing subscan %s to %s", name, url)
        elif not self._nxentry_created:
            self._exception_is_fatal = True
            raise RuntimeError(
                f"Scan {name} already exists in {self._scan_info.filename}"
            )
        return nxentry

    @contextmanager
    def nxinstrument(
        self, subscan: info.SubScanInfo
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxentry(subscan) as nxentry:
            if nxentry is None:
                yield None
            else:
                yield nexus.nxInstrument(
                    nxentry, "instrument", nxdict=self._scan_info.instrument_info
                )

    @contextmanager
    def nxdetector(
        self, subscan: info.SubScanInfo, name: str, **kwargs
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxinstrument(subscan) as nxinstrument:
            if nxinstrument is None:
                yield None
            else:
                yield nexus.nxDetector(nxinstrument, name, **kwargs)

    @contextmanager
    def nxpositioner(
        self, subscan: info.SubScanInfo, name: str, **kwargs
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxinstrument(subscan) as nxinstrument:
            if nxinstrument is None:
                yield None
            else:
                yield nexus.nxPositioner(nxinstrument, name, **kwargs)

    @contextmanager
    def nxcollection(
        self, subscan: info.SubScanInfo, name: str, **kwargs
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxinstrument(subscan) as nxinstrument:
            if nxinstrument is None:
                yield None
            else:
                yield nexus.nxCollection(nxinstrument, name, **kwargs)

    @contextmanager
    def nxpositioners(
        self, subscan: info.SubScanInfo, suffix: str = ""
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxcollection(subscan, "positioners" + suffix) as nxcollection:
            yield nxcollection

    @contextmanager
    def nxmeasurement(
        self, subscan: info.SubScanInfo
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxentry(subscan) as nxentry:
            if nxentry is None:
                yield None
            else:
                yield nexus.nxCollection(nxentry, "measurement")

    @contextmanager
    def nxnotes(
        self, subscan: info.SubScanInfo
    ) -> Generator[Optional[h5py.Group], None, None]:
        with self.nxentry(subscan) as nxentry:
            if nxentry is None:
                yield None
            else:
                yield nexus.nxCollection(nxentry, "notes")

    @contextmanager
    def device_parent_context(
        self, subscan, device: Dict
    ) -> Generator[Optional[h5py.Group], None, None]:
        parentname = normalize_nexus_name(device["device_name"])
        if device["device_type"] in ("positioner", "positionergroup"):
            # Add as separate positioner group
            parentcontext = self.nxpositioner
            parentcontextargs = subscan, parentname
        elif device["device_name"]:
            # Add as separate detector group
            parentcontext = self.nxdetector
            parentcontextargs = subscan, parentname
        else:
            # Add to generic 'measurement' group
            parentcontext = self.nxmeasurement
            parentcontextargs = (subscan,)
        with parentcontext(*parentcontextargs) as parent:
            yield parent

    @contextmanager
    def instrument_group_context(
        self, subscan: info.SubScanInfo, groupname: str
    ) -> Generator[Optional[h5py.Group], None, None]:
        parentcontext = None
        for device in self._scan_info.devices[subscan.name].values():
            parentname = normalize_nexus_name(device["device_name"])
            if groupname != parentname:
                continue
            if device["device_type"] in ("positioner", "positionergroup"):
                parentcontext = self.nxpositioners
            elif device["device_name"]:
                parentcontext = self.nxdetector
            break
        if parentcontext is None:
            yield None
            return
        with parentcontext(subscan, groupname) as parent:
            yield parent

    def _save_reference_mode(self, file_format):
        """
        Save reference as external link (or string uri) or copy data to hdf5
        """
        if file_format == "hdf5":
            external = self._allow_external_hdf5 and nexus.HASVIRTUAL
        elif file_format == "edf":
            external = self._allow_external_nonhdf5
        else:
            external = True
            file_format = None
        if not external and not self._copy_non_external:
            file_format = None
            external = True
        return external, file_format

    def _save_device_metadata(self, subscan: info.SubScanInfo):
        """Get all device metadata and save them under the appropriate NXdetector or NXpositioner groups."""
        channels = self._scan_info.get("channels")
        devices = self._scan_info.get("devices")
        visited = set()
        for fullname, device in self._scan_info.devices[subscan.name].items():
            if device["device_name"] in visited:
                continue
            channel_info = channels.get(fullname)
            if not channel_info:
                continue
            device_name = channel_info.get("device")
            if not device_name:
                continue
            metadata = devices[device_name].get("metadata")
            if not metadata:
                continue

            metadata_keys = device["metadata_keys"]
            if metadata_keys:
                metadata = {
                    metadata_keys[k]: v
                    for k, v in metadata.items()
                    if k in metadata_keys
                }
            if not metadata:
                continue

            with self.device_parent_context(subscan, device) as parent:
                if parent is None:
                    continue
                if not parent.name.endswith("measurement"):
                    self._device_dicttonx_helper(parent, metadata)
                visited.add(device["device_name"])

    def _update_subscan_status(self, subscan: info.SubScanInfo, status: str) -> None:
        status = status.upper()
        final = status in ("SUCCEEDED", "FAILED")
        with self.nxentry(subscan) as nxentry:
            if nxentry is None:
                return
            nxnote = nexus.nxNote(nxentry, "writer")
            nexus.updateDataset(nxnote, "status", status)
            with self._modify_nxroot():
                nexus.updated(nxentry, final=final, parents=True)
            subscan.logger.info("Scan marked as %s in HDF5", status)

    def positioner_iter(
        self,
        subscan: info.SubScanInfo,
        onlyprincipals: bool = True,
        onlymasters: bool = True,
    ) -> Iterator[Tuple[str, datasets.Hdf5DatasetWriter]]:
        """Yields all positioner dataset handles

        :param subscan:
        :param onlyprincipals: only the principal value of each positioner
        :param onlymasters: only positioners that are master in the acquisition chain
        :returns: fullname and dataset handles
        """
        for fullname, dproxy in list(subscan.dataset_items()):
            if dproxy.device_type in ("positioner", "positionergroup"):
                if onlyprincipals and dproxy.data_type != "principal":
                    continue
                if onlymasters and dproxy.master_index < 0:
                    continue
                yield fullname, dproxy

    def _iter_detector_datasets(
        self, subscan: info.SubScanInfo
    ) -> Iterator[Tuple[str, datasets.Hdf5DatasetWriter]]:
        """Yields all dataset handles except for positioners

        :param subscan:
        :returns: fullname and dataset handle
        """
        for fullname, dproxy in list(subscan.dataset_items()):
            if dproxy.device_type not in ("positioner", "positionergroup"):
                yield fullname, dproxy

    def _save_positioners(self, subscan: info.SubScanInfo) -> None:
        """Save fixed snapshots of motor positions.

        :param Subscan subscan:
        """
        info = self._scan_info.positioner_info
        units = info.get("positioners_units", {})

        # Positions at the beginning of the scan
        positioners = info.get("positioners_start", {})
        subscan.logger.info("Save {} motor positions".format(len(positioners)))
        self._save_positioners_snapshot(
            subscan, positioners, units, "_start", overwrite=False
        )
        self._save_positioners_snapshot(
            subscan, positioners, units, "", overwrite=False
        )
        positioners = info.get("positioners_dial_start", {})
        self._save_positioners_snapshot(
            subscan, positioners, units, "_dial_start", overwrite=False
        )

        # Positions at the end of the scan
        positioners = info.get("positioners_end", {})
        self._save_positioners_snapshot(
            subscan, positioners, units, "_end", overwrite=True
        )
        positioners = info.get("positioners_dial_end", {})
        self._save_positioners_snapshot(
            subscan, positioners, units, "_dial_end", overwrite=True
        )

    def _save_positioners_snapshot(
        self,
        subscan: info.SubScanInfo,
        positions: Mapping,
        units: Mapping,
        suffix: str,
        overwrite: bool = False,
    ) -> None:
        """Save fixed snapshot of motor positions.

        :param subscan:
        :param positions: name:position
        :param units: name:unit
        :param suffix: output suffix
        :param overwrite: goes for values and attributes
        """
        if not positions:
            return
        with self.nxpositioners(subscan, suffix=suffix) as nxpositioners:
            if nxpositioners is None:
                return
            for mot, pos in positions.items():
                unit = units.get(mot, None)
                exists = mot in nxpositioners
                if exists:
                    dset = nxpositioners[mot]
                    if overwrite:
                        dset[()] = pos
                    if unit and ("units" not in dset.attrs or overwrite):
                        dset.attrs["units"] = unit
                else:
                    if unit:
                        attrs = {"units": unit}
                    else:
                        attrs = {}
                    nexus.nxCreateDataSet(nxpositioners, mot, pos, attrs)

    def _add_to_measurement_group(
        self, subscan: info.SubScanInfo, dproxy: datasets.Hdf5DatasetWriter
    ) -> None:
        """Add link in measurement group."""
        with self.nxmeasurement(subscan) as measurement:
            if measurement is None:
                return
            if dproxy.parent == measurement.name:
                return
            linkname = dproxy.linkname
            if not linkname:
                dproxy.logger.warning("cannot be linked too")
                return
            linknames = [linkname]
            for linkname in linknames:
                if linkname in measurement:
                    self._scan_info.logger.warning(
                        "Duplicate name '%s' in the measurement group. Rename this detector or positioner.",
                        linkname,
                    )
                else:
                    nexus.createLink(measurement, linkname, dproxy.path)

    def _add_to_positioners_group(
        self, subscan: info.SubScanInfo, dproxy: datasets.Hdf5DatasetWriter
    ) -> None:
        """Add link in positioners group."""
        with self.nxpositioners(subscan) as parent:
            if parent is None:
                return
            linkname = dproxy.linkname
            try:
                del parent[linkname]
            except KeyError:
                pass
            nexus.createLink(parent, linkname, dproxy.path)

    def _save_subscan_user_metadata(self, subscan: info.SubScanInfo) -> None:
        """Dump metadata under the user scan metadata categories"""
        subscan.logger.info("Save scan user metadata")
        categories = set(self._scan_info["scan_meta_categories"])
        categories -= {"positioners", "nexuswriter"}
        nxtreedict = {cat: self._scan_info.get(cat, None) for cat in categories}
        self._subscan_dicttonx(subscan, nxtreedict)

    def _save_subscan_metadata(self, subscan: info.SubScanInfo) -> None:
        """Dump metadata dataset metadata"""
        subscan.logger.info("Save scan metadata")
        if self._scan_info.get("data_policy", "").upper() != "ESRF":
            return
        metadict = self._scan_info.get("dataset_metadata_snapshot")
        if not metadict:
            return
        # Remove unwanted ICAT field names (see hdf5_cfg.xml)
        skip = ("SamplePositioners_value", "SamplePositioners_name")
        metadict = {
            k: v
            for k, v in metadict.items()
            if k.startswith("Sample") and k not in skip
        }
        if not metadict:
            return
        nxtreedict = create_nxtreedict(metadict)
        self._subscan_dicttonx(subscan, nxtreedict)

    def _save_scan_status(self, subscan: info.SubScanInfo) -> None:
        """Save the reason the scan ended in Bliss"""
        end_reason = self._scan_info.get("end_reason", "NOT STARTED")
        with self.nxentry(subscan) as parent:
            if parent is None:
                return
            parent["end_reason"] = end_reason

    def _subscan_dicttonx(
        self, subscan: info.SubScanInfo, nxtreedict: Optional[Mapping]
    ) -> None:
        if not nxtreedict:
            return
        with self.nxentry(subscan) as parent:
            if parent is None:
                return

            nxtreedict_direct = dict()

            for catname, catcontent in nxtreedict.items():
                if not catcontent:
                    continue
                # Datasets are dumped directly
                if not isinstance(catcontent, Mapping):
                    nxtreedict_direct[catname] = catcontent
                    continue

                # Empty group
                if "NX_class" in catcontent:
                    catcontent["@NX_class"] = catcontent.pop("NX_class")
                if not (set(catcontent.keys()) - {"@NX_class"}):
                    continue

                # Instrument group should not be dumped directly
                if catname == "instrument":
                    directcontent = dict()
                    for name, value in catcontent.items():
                        # Datasets are dumped directly
                        if not isinstance(value, Mapping):
                            directcontent[name] = value
                            continue
                        with self.instrument_group_context(subscan, name) as group:
                            if group is None:
                                # Non-device groups are dumped directly
                                directcontent[name] = value
                                continue
                            # Device group is dumped with a specific parent for
                            # initialization purposes
                            self._device_dicttonx_helper(group, value)
                            continue
                    if not directcontent:
                        continue
                    catcontent = directcontent

                # Group can be dumped directly
                nxtreedict_direct[catname] = catcontent

            self._subscan_dicttonx_helper(subscan, parent, nxtreedict_direct)

    def _subscan_dicttonx_helper(
        self,
        subscan: info.SubScanInfo,
        parent: h5py.Group,
        nxtreedict: Optional[Mapping],
    ) -> None:
        if not nxtreedict or not purge_nxtreedict(nxtreedict):
            return
        try:
            nexus.dicttonx(nxtreedict, parent, update_mode="modify", add_nx_class=True)
        except Exception as e:
            self._set_state(self.STATES.FAULT, e)
            subscan.logger.exception(
                "Scan metadata cannot be saved:\n%s", pformat(nxtreedict)
            )

    def _device_dicttonx_helper(
        self, parent: h5py.Group, nxtreedict: Optional[Mapping]
    ):
        if not nxtreedict or not purge_nxtreedict(nxtreedict):
            return
        nxtreedict["@NX_class"] = parent.attrs["NX_class"]
        nexus.dicttonx(nxtreedict, parent, update_mode="modify", add_nx_class=True)

    def _save_subscan_notes(self, subscan: info.SubScanInfo) -> None:
        """Save notes for this subscan"""
        notes = self._scan_info.get("comments", [])
        if not notes:
            return
        with self.nxnotes(subscan) as parent:
            if parent is None:
                return
            subscan.logger.info("Save scan notes")
            for i, note in enumerate(notes, 1):
                nexus.nxNote(
                    parent,
                    f"note_{i:02d}",
                    data=note["message"],
                    type="text/plain",
                    date=datetime.datetime.fromisoformat(note["date"]),
                )

    def _create_master_links(self, subscan: info.SubScanInfo) -> None:
        """Links to the scan's NXentry"""
        filenames = self._scan_info.get_master_filenames()
        if not filenames:
            return
        with self.nxentry(subscan) as nxentry:
            if nxentry is None:
                return
            self._scan_info.logger.info("Create scan links in masters ...")
            prefix, _ = os.path.splitext(os.path.basename(nxentry.file.filename))
            prefix += "_"
            nxentry_name = nxentry.name[1:]  # remove the leading "/"
            for level, filename in filenames.items():
                if level == "dataset":
                    linkname = nxentry_name
                else:
                    linkname = prefix + nxentry_name
                with self.nxroot(filename=filename) as nxroot:
                    if nxroot is None:
                        continue
                    if linkname in nxroot:
                        continue
                    self._scan_info.logger.info(
                        "Create link '%s' in master '%s'", linkname, filename
                    )
                    nexus.createLink(nxroot, linkname, nxentry)

    def _create_plots(self, subscan: info.SubScanInfo) -> None:
        """Create default plot in Nexus structure"""
        with self.nxentry(subscan) as nxentry:
            if nxentry is None:
                return
            default = None
            plots = self._scan_info.plots
            if plots:
                positioners = [
                    dproxy
                    for _, dproxy in self.positioner_iter(
                        subscan, onlyprincipals=True, onlymasters=True
                    )
                ]
                plotselect = self._scan_info.plotselect
                subscan.logger.info(f"Create {len(plots)} plots")
            else:
                subscan.logger.info("No plots defined for saving")
            for plotname, plotparams in plots.items():
                if plotname in nxentry:
                    subscan.logger.warning(
                        f"Cannot create plot {repr(plotname)} (name already exists)"
                    )
                    continue
                nxproxy = self._create_nxdata_proxy(subscan, plotname, **plotparams)
                if not nxproxy:
                    nxproxy.logger.warning("Not created (no signals)")
                    continue
                nxproxy.add_axes(positioners, self._scan_info._save_order)
                default = nxproxy.save(nxentry, default, plotselect)
            # Default plot
            with self._modify_nxroot():
                if default is None:
                    nexus.markDefault(nxentry)
                else:
                    nexus.markDefault(default)

    def _create_nxdata_proxy(
        self,
        subscan: info.SubScanInfo,
        plotname: str,
        ndim: int = -1,
        grid: bool = False,
        items: List = None,
    ) -> plots.NXdataWriter:
        """Select plot signals based on detector dimensions.

        :param subscan:
        :param plotname:
        :param ndim: detector dimensions
        :param grid: preserve scan shape
        :returns:
        """
        nxproxy = plots.NXdataWriter(plotname, parent_logger=subscan.logger)
        if not items and ndim < 0:
            return nxproxy
        detector_datasets = sorted(
            self._iter_detector_datasets(subscan),
            key=lambda tpl: tpl[1].linkname or tpl[1].name,
        )
        if items:
            for configname in items:
                for fullname, dproxy in detector_datasets:
                    if self._matching_fullname(configname, fullname):
                        nxproxy.add_signal(dproxy, grid)
                        break
        else:
            for _, dproxy in detector_datasets:
                if dproxy.detector_ndim == ndim:
                    nxproxy.add_signal(dproxy, grid)
        return nxproxy

    @contextmanager
    def nxapplication(
        self, subscan: info.SubScanInfo, name: str, definition_name: str
    ) -> Generator[Optional[h5py.Group], None, None]:
        """Yields the NXsubentry instance (h5py.Group) or None when NXentry is missing

        :param subscan:
        :param name: name of the NXsubentry
        :param definition_name: Nexus application definition
        """
        with self.nxentry(subscan) as nxentry:
            if nxentry is None:
                yield None
                return
            # Find nxapplication (when existing)
            for child in nxentry:
                nxsubentry = nxentry[child]
                if (
                    child == name
                    and nexus.isNxClass(nxsubentry, "NXsubentry")
                    and definition_name == nxsubentry.attrs.get("definition", None)
                ):
                    break
            else:
                nxsubentry = None
            if nxsubentry is None:
                # Create nxapplication
                ret = self._nxapplication_create_args(name, definition_name)
                if ret is None:
                    yield None
                    return
                args, kwargs = ret
                nxsubentry = nexus.nxSubEntry(nxentry, *args, **kwargs)
            yield nxsubentry

    def _nxapplication_create_args(
        self, name: str, definition_name: str
    ) -> Tuple[tuple, dict]:
        """Arguments for application nxSubEntry creation"""
        start_time = self._scan_info.get("start_time")
        if not start_time:
            self._scan_info._log_h5creation("'start_time' not specified")
            return None
        start_time = datetime.datetime.fromisoformat(start_time)
        nxdict = {"definition": definition_name}
        args = (name,)
        kwargs = {"start_time": start_time, "nxdict": nxdict}
        return args, kwargs

    def _save_applications(self, subscan: info.SubScanInfo):
        """Save a Nexus application definitions"""
        for name, nxapplidef in self._scan_info.applications.items():
            nxclass = nxapplidef["class"].lower()
            save_app = self._applications.get(nxclass)
            if save_app is None:
                self._scan_info.logger.warning(
                    "NeXus application definitions '%s' is not supported", nxclass
                )
            else:
                save_app(subscan, name, nxapplidef)

    def _save_application_xrf(
        self, subscan: info.SubScanInfo, name: str, xrfapplidef: Mapping
    ):
        """Save XRF Nexus application definition"""
        with self.nxapplication(subscan, name, "APPxrf") as nxsubentry:
            if nxsubentry is None:
                return
            self._scan_info.logger.info("Create 'APPxrf' Nexus application")
            self._save_application_i0(subscan, nxsubentry, xrfapplidef)
            self._save_application_it(subscan, nxsubentry, xrfapplidef)
            self._save_application_mca(subscan, nxsubentry, xrfapplidef)
            nexus.updated(nxsubentry, final=True, parents=False)

    def _mca_iter(
        self, subscan: info.SubScanInfo
    ) -> Iterator[Tuple[str, datasets.Hdf5DatasetWriter]]:
        """Yields all principal mca dataset handle

        :param subscan:
        :returns: fullname and dataset handle
        """
        for fullname, dproxy in list(subscan.dataset_items()):
            if dproxy.device_type == "mca" and dproxy.data_type == "principal":
                yield fullname, dproxy

    def _get_application(
        self, nxapplidef: Mapping, attr: str, default: Any = None
    ) -> Any:
        """Get parameter from application definition"""
        ret = nxapplidef.get(attr, None)
        if ret is None:
            self._scan_info.logger.warning(
                "Application definition incomplete ('%s' is missing)", attr
            )
            ret = default
        return ret

    def _save_application_i0(
        self, subscan: info.SubScanInfo, parent: h5py.Group, nxapplidef: Mapping
    ) -> None:
        """Add beam monitor to the application

        :param parent: application subentry
        :param nxapplidef:
        """
        I0configname = self._get_application(nxapplidef, "I0")
        I0appliname = "i0"
        self._save_application_link(subscan, parent, I0configname, I0appliname)

    def _save_application_it(
        self, subscan: info.SubScanInfo, parent: h5py.Group, nxapplidef: Mapping
    ) -> None:
        """Add sample transmission to the application

        :param h5py.Group parent: application subentry
        :param dict nxapplidef:
        """
        Itconfigname = self._get_application(nxapplidef, "It")
        Itappliname = "it"
        self._save_application_link(subscan, parent, Itconfigname, Itappliname)

    def _save_application_mca(
        self, subscan: info.SubScanInfo, parent: h5py.Group, xrfapplidef: Mapping
    ) -> None:
        """Add MCA's tp the application

        :param Subscan subscan:
        :param h5py.Group parent: application subentry
        :param dict xrfapplidef:
        """
        # Mca names from the beamline configuration
        confignames = self._get_application(xrfapplidef, "mca", [])
        if not confignames:
            # All MCA's when not specified
            for fullname, dproxy in self._mca_iter(subscan):
                confignames.append(fullname)
            confignames = sorted(confignames)
        if not confignames:
            self._scan_info.logger.warning(
                "Application definition incomplete (no mca's found)"
            )
            return

        # Links or concatenate?
        concatenate = self._stack_mcas
        aslinks = len(confignames) == 1 or not concatenate
        if aslinks:
            method = self._save_application_links
        else:
            method = self._save_application_merged

        # Add mca
        withsuffix = len(confignames) > 1 and not concatenate
        if withsuffix:
            mcaapplinamefmt = "mca{:02d}"
        else:
            mcaapplinamefmt = "mca"
        method(
            subscan,
            parent,
            confignames,
            mcaapplinamefmt,
            devicetype="mca",
            datatype="principal",
        )

        if withsuffix:
            mcaapplinamefmt = "elapsed_time_mca{:02d}"
        else:
            mcaapplinamefmt = "elapsed_time"
        method(
            subscan,
            parent,
            confignames,
            mcaapplinamefmt,
            devicetype="mca",
            datatype="realtime",
        )

        if withsuffix:
            mcaapplinamefmt = "live_time_mca{:02d}"
        else:
            mcaapplinamefmt = "live_time"
        method(
            subscan,
            parent,
            confignames,
            mcaapplinamefmt,
            devicetype="mca",
            datatype="livetime",
        )

    def _save_application_merged(
        self,
        subscan: info.SubScanInfo,
        parent: h5py.Group,
        confignames: List[str],
        appliname: str,
        virtual: bool = True,
        **kwargs,
    ) -> None:
        """Add datasets to an application as one single dataset,
        virtual or merged copy.

        :param subscan:
        :param parent: application subentry
        :param confignames: names specified in the beamline
                            static configuration
        :param applifmt: format for application definition name
        :param virtual: virtual dataset or copy
        :param **kwargs: see `_iter_fullnames`
        """
        if not confignames or appliname in parent:
            return
        uris = []
        for configname in confignames:
            notfoundmsg = (
                "Application definition incomplete ({} not found for {})".format(
                    repr(configname), repr(appliname)
                )
            )
            for fullname in self._iter_fullnames(
                subscan, configname, notfoundmsg=notfoundmsg, **kwargs
            ):
                dproxy = subscan.get_dataset_writer(fullname)
                with dproxy.open() as dset:
                    if not uris:
                        value = {
                            "data": uris,
                            "fillvalue": dset.fillvalue,
                            "axis": 0,
                            "newaxis": True,
                            "virtual": virtual,
                        }
                        attrs = dset.attrs
                    uris.append(nexus.getUri(dset))
        if uris:
            nexus.nxCreateDataSet(parent, appliname, value, attrs)

    def _save_application_links(
        self,
        subscan: info.SubScanInfo,
        parent: h5py.Group,
        confignames: List[str],
        applifmt: str,
        **kwargs,
    ) -> None:
        """Add dataset links to an application.

        :param subscan:
        :param parent: application subentry
        :param confignames: names specified in the beamline
                            static configuration
        :param applifmt: format for application definition name
        :param **kwargs: see `_iter_fullnames`
        """
        if not confignames:
            return
        for i, configname in enumerate(confignames):
            appliname = applifmt.format(i)
            self._save_application_link(
                subscan, parent, configname, appliname, **kwargs
            )

    def _save_application_link(
        self,
        subscan: info.SubScanInfo,
        parent: h5py.Group,
        configname: str,
        appliname: str,
        **kwargs,
    ) -> None:
        """Link to a dataset in the application definition

        :param subscan:
        :param parent: application subentry
        :param configname: name specified in the beamline
                               static configuration
        :param appliname: name in the Nexus application definition
        :param **kwargs: see `_iter_fullnames`
        """
        if not configname or appliname in parent:
            return
        notfoundmsg = "Application definition incomplete ({} not found for {})".format(
            repr(configname), repr(appliname)
        )
        for fullname in self._iter_fullnames(
            subscan, configname, notfoundmsg=notfoundmsg, **kwargs
        ):
            dproxy = subscan.get_dataset_writer(fullname)
            nexus.createLink(parent, appliname, dproxy.path)

    def _matching_fullname(self, configname: str, fullname: str) -> bool:
        """Checks whether a Redis node's full name is referred to
        by name from the writer configuration.

        Examples:
            "iodet" refers to "simulation_diode_controller:iodet"
            "xmap1:det0" refers to "xmap1:realtime_det0"
            "xmap1:det0" refers to "simxmap1:spectrum_det0"

        :param configname: from the writer configuration
        :param fullname: node.fullname
        """
        seps = r"[\.:]"
        configparts = re.split(seps, configname)
        fullparts = re.split(seps, fullname)
        return all(
            pfull.endswith(pconfig)
            for pfull, pconfig in zip(fullparts[::-1], configparts[::-1])
        )

    def _iter_fullnames(
        self,
        subscan: info.SubScanInfo,
        configname: str,
        devicetype: Optional[str] = None,
        datatype: Optional[str] = None,
        notfoundmsg: Optional[str] = None,
    ) -> Iterator[str]:
        """Yield all Redis node's full names referred to by a name
        from the writer configuration.

        :param Subscan subscan:
        :param configname: name specified in the beamline
                           static configuration
        :param devicetype: device type
        :param datatype: data type
        :param notfoundmsg:
        :yields: Redis node fullname
        """
        incomplete = True
        for fullname, dproxy in self._iter_detector_datasets(subscan):
            if self._matching_fullname(configname, fullname):
                if (devicetype == dproxy.device_type or not devicetype) and (
                    datatype == dproxy.data_type or not datatype
                ):
                    incomplete = False
                    yield fullname
        if incomplete and notfoundmsg:
            self._scan_info.logger.warning(notfoundmsg)


def purge_nxtreedict(nxtreedict: dict) -> bool:
    """Returns `True` when the tree contains actual content."""
    for key, value in list(nxtreedict.items()):
        if isinstance(value, dict):
            has_content = purge_nxtreedict(value)
            if not has_content:
                nxtreedict.pop(key)
    return bool(set(nxtreedict.keys()) - {"@NX_class"})
